import os
import time
import gc
from pathlib import Path

from cordex.representations.lookup import LookupLexicon, SUPPORTED_LOOKUP_LANGUAGES, LookupApi
from cordex.utils.progress_bar import progress
from cordex.structures.syntactic_structure import build_structures
from cordex.matcher.match_store import MatchStore
from cordex.statistics.word_stats import WordStats
from cordex.writers.formatter import OutFormatter, OutNoStatFormatter
from cordex.writers.writer import Writer
from cordex.readers.loader import load_files
from cordex.database.database import Database
from cordex.utils.time_info import TimeInfo

from cordex.postprocessors.postprocessor import Postprocessor
import logging

logger = logging.getLogger('cordex')

HOME_DIR = str(Path.home())

class Pipeline:
    def __init__(self, structures, **kwargs):
        kwargs['structures'] = structures
        self.args = self.set_default_args(kwargs)

        self.structures, self.max_num_components, is_ud = build_structures(self.args)
        self.args['is_ud'] = is_ud

        if self.args['lookup_api'] and not is_ud:
            self.lookup_api = LookupApi('https://blisk.ijs.si/api')
            self.lookup_lexicon = None
        elif self.args['lookup_lexicon'] is not None and os.path.exists(self.args['lookup_lexicon']) and not is_ud:
            self.lookup_lexicon = LookupLexicon(self.args['lookup_lexicon'])
            self.lookup_api = None
        else:
            self.lookup_lexicon = None
            self.lookup_api = None

        if not self.lookup_lexicon and self.args['lang'] in SUPPORTED_LOOKUP_LANGUAGES and not is_ud:
            logger.warning(f'WARNING: Results could be improved if you include lookup lexicon (or provide correct path '
                           f'to it using `lookup_lexicon` argument). '
                           f'You may download it using `cordex.download()`. If it is stored in different location than '
                           f'`{HOME_DIR}/cordex_resources` you should add path to `lookup_lexicon` argument.')

    def __call__(self, corpus):
        if type(corpus) == str:
            corpus = [corpus]
        self.args['corpus'] = corpus
        time_info = TimeInfo(len(self.args['corpus']))

        database = Database(self.args)
        self.match_store = MatchStore(self.args, database)
        self.word_stats = WordStats(self.args, database)
        postprocessor = Postprocessor(fixed_restriction_order=self.args['fixed_restriction_order'], lang=self.args['lang'])

        for words in load_files(self.args, database):
            if words is None:
                time_info.add_measurement(-1)
                continue

            start_time = time.time()
            matches = self.match_file(words, self.structures, postprocessor)

            # adds results to database
            self.match_store.add_matches(matches)
            self.word_stats.add_words(words)
            database.commit()

            # force a bit of garbage collection
            del words
            del matches
            gc.collect()

            time_info.add_measurement(time.time() - start_time)
            time_info.info()

        # get word renders for lemma/msd
        self.word_stats.lowercase_words_under_threshold()
        self.word_stats.generate_renders()
        self.match_store.determine_collocation_dispersions()

        # figure out representations!
        self.match_store.set_representations(self.word_stats, self.structures, self.args['is_ud'], lookup_lexicon=self.lookup_lexicon, lookup_api=self.lookup_api)

        return self

    def write(self, path, separator='\t', sort_by=-1, sort_reversed=False, decimal_separator='.'):
        self.args['out'] = path
        self.args['separator'] = separator
        self.args['decimal_separator'] = decimal_separator
        self.args['sort_by'] = sort_by
        self.args['sort_reversed'] = sort_reversed
        self.args['multiple_output'] = '.' not in Path(path).name

        # if no output files, just exit
        if all([x is None for x in [self.args['out']]]):
            return

        if self.args['statistics']:
            Writer.make_output_writer(self.args, self.max_num_components, self.match_store, self.word_stats,
                                      self.args['is_ud']).write_out(
                self.structures, self.match_store)
        else:
            Writer.make_output_no_stat_writer(self.args, self.max_num_components, self.match_store, self.word_stats,
                                              self.args['is_ud']).write_out(
                self.structures, self.match_store)

    def get_list(self, separator='\t', sort_by=-1, sort_reversed=False, decimal_separator='.'):
        self.args['separator'] = separator
        self.args['decimal_separator'] = decimal_separator
        self.args['sort_by'] = sort_by
        self.args['sort_reversed'] = sort_reversed
        self.args['multiple_output'] = False

        params = Writer.other_params(self.args)
        if self.args['statistics']:
            writer = Writer(None, self.max_num_components, OutFormatter(self.match_store, self.word_stats, self.args['is_ud'], self.args),
                            self.args['collocation_sentence_map_dest'], params, self.args['separator'])
        else:
            writer = Writer(None, self.max_num_components,
                            OutNoStatFormatter(self.match_store, self.word_stats, self.args['is_ud'], self.args),
                            self.args['collocation_sentence_map_dest'], params, self.args['separator'])
        return writer.write_out(self.structures, self.match_store, return_list=True)

    @staticmethod
    def match_file(words, structures, postprocessor):
        """ Looks for collocations inside a file that match structure restrictions. """
        matches = {s: [] for s in structures}

        for s in progress(structures, "matching"):
            for w in words:
                mhere = s.match(w)
                for match in mhere:
                    if not postprocessor.is_fixed_restriction_order(match):
                        continue
                    collocation_id = [[idx, w.lemma] for idx, w in match.items()]
                    collocation_id = [s.id] + list(sorted(collocation_id, key=lambda x: x[0]))
                    match, collocation_id = postprocessor.process(match, collocation_id)
                    collocation_id = tuple(collocation_id)

                    matches[s].append((match, collocation_id))

        return matches

    @staticmethod
    def set_default_args(kwargs):
        """ Sets default arguments. """
        lang = 'sl'
        default_args = {
            'min_freq': 0,
            'db': None,
            'overwrite_db': False,
            'jos_msd_lang': 'en',
            'ignore_punctuations': False,
            'fixed_restriction_order': False,
            'lookup_lexicon': f'{HOME_DIR}/cordex_resources/{lang}.xz',
            'lookup_api': False,
            'statistics': True,
            'lang': 'sl',
            'collocation_sentence_map_dest': None,
            'jos_depparse_lang': 'sl'
        }

        return {**default_args, **kwargs}



"""
A file that takes care of order of children in structures.
"""
from enum import Enum


class Order(Enum):
    FromTo = 0
    ToFrom = 1
    Any = 2

    @staticmethod
    def new(order):
        """ Creates a new order. """
        if order is None:
            return Order.Any
        elif order == "to-from":
            return Order.ToFrom
        elif order == "from-to":
            return Order.FromTo
        else:
            raise NotImplementedError("What kind of ordering is: {}".format(order))

    def match(self, from_w, to_w):
        """ Checks whether word order is correct. """
        if self is Order.Any:
            return True

        fi = from_w.int_id
        ti = to_w.int_id

        if self is Order.FromTo:
            return fi < ti
        elif self is Order.ToFrom:
            return ti < fi
        else:
            raise NotImplementedError("Should not be here: Order match")



"""
Class for loading matches
"""
from ast import literal_eval

from cordex.words.word import WordUD, WordJOS


class StructureMatch:
    def __init__(self, match_id, structure):
        self.match_id = str(match_id)
        self.structure = structure

        self.matches = []
        self.representations = {}
    
    @staticmethod
    def from_db(db, collocation_id, structure, is_ud):
        """ Loads matches from database. """
        result = StructureMatch(collocation_id, structure)
        prev_match_id = None

        if is_ud:
            stmt = """SELECT match_id, component_id, word_lemma, word_text, word_udpos, word_id, sentence_id
                      FROM CollocationMatches 
                      JOIN Matches ON Matches.match_id=CollocationMatches.mid_match_id 
                      WHERE mid_collocation_id=? 
                      ORDER BY match_id"""

            for row in db.execute(stmt, (collocation_id,)):
                match_id, component_id, word_lemma, word_text, word_udpos, word_id, sentence_id = row

                if match_id != prev_match_id:
                    result.matches.append({})
                    prev_match_id = match_id

                word_udpos = literal_eval(word_udpos)
                int_word_id = int(word_id) if word_id[0] in '0123456789' else int(word_id[1:])
                result.matches[-1][str(component_id)] = WordUD(word_lemma, '', sentence_id, word_id, int_word_id, word_text, False, feats=word_udpos)

        else:
            stmt = """SELECT match_id, component_id, word_lemma, word_text, word_xpos, word_id, sentence_id
                      FROM CollocationMatches 
                      JOIN Matches ON Matches.match_id=CollocationMatches.mid_match_id 
                      WHERE mid_collocation_id=? 
                      ORDER BY match_id"""

            for row in db.execute(stmt, (collocation_id,)):
                match_id, component_id, word_lemma, word_text, word_xpos, word_id, sentence_id = row

                if match_id != prev_match_id:
                    result.matches.append({})
                    prev_match_id = match_id

                int_word_id = int(word_id) if word_id[0] in '0123456789' else int(word_id[1:])
                result.matches[-1][str(component_id)] = WordJOS(word_lemma, word_xpos, sentence_id, word_id, int_word_id, word_text, False, False)
        
        for component_id, text, msd in db.execute("SELECT component_id, text, msd FROM Representations WHERE collocation_id=?", (collocation_id,)):
            result.representations[str(component_id)] = (text, msd)
        
        return result

    def distinct_forms(self):
        """ Counts all distinct forms. """
        dm = set()
        keys = list(self.matches[0].keys())
        for words in self.matches:
            dm.add(" ".join(words[k].text for k in keys))
        return len(dm)

    def __len__(self):
        return len(self.matches)



"""
Default syntactic structure class.
"""

from xml.etree import ElementTree
import logging

from cordex.utils.codes_tagset import PPB_DEPRELS
from cordex.structures.component import Component, ComponentType


class SyntacticStructure:
    def __init__(self):
        self.id = None
        # self.lbs = None
        self.components = []
        self.fake_root_included = False

    @staticmethod
    def from_xml(xml, no_stats, is_ud):
        """ Reads a syntactic structure from xml. """
        st = SyntacticStructure()
        st.id = xml.get('id')
        if st.id is None:
            st.id = xml.get('tempId')

        components, dependencies, definitions = list(xml)

        deps = [(dep.get('from'), dep.get('to'), dep.get('label'), dep.get('order'))
                for dep in dependencies]
        comps = {comp.get('cid'): dict(comp.items()) for comp in components}

        restrs, forms = {}, {}

        for comp in definitions:
            n = comp.get('cid')
            restrs[n] = []
            forms[n] = []

            for el in comp:
                if el.tag.startswith("restriction"):
                    restrs[n].append(el)
                elif el.tag.startswith("representation"):
                    st.add_representation(n, el, forms)
                else:
                    raise NotImplementedError("Unknown definition: {} in structure {}"
                                              .format(el.tag, st.id))

        # creates fake root component onto which other components are appended.
        fake_root_component = Component({'cid': '#', 'type': 'other', 'restriction': None}, is_ud)
        fake_root_component_children = fake_root_component.create_children(deps, comps, restrs, forms)

        # all dep with value modra point to artificial root - fake_root_component
        if any([dep[2] == 'modra' for dep in deps]):
            st.fake_root_included = True
            st.components = [fake_root_component] + fake_root_component_children
        else:
            st.components = fake_root_component_children

        if not no_stats:
            if is_ud:
                st.determine_core2w_ud()
            else:
                st.determine_core2w()

        return st

    def determine_core2w_ud(self):
        """ Determines two core words in UD collocation. """
        deprels = {}
        for c in self.components:
            for next_el in c.children:
                deprels[next_el[0]] = next_el[1]
        ppb_components_num = 0
        for c in self.components:
            if c.type != ComponentType.Core:
                continue
            if c in deprels and deprels[c] not in PPB_DEPRELS:
                continue
            ppb_components_num += 1
            c.type = ComponentType.Core2w

        assert ppb_components_num == 2, RuntimeError("Cannot determine 2 'jedrna polnopomenska beseda' for", self.id)

    def determine_core2w(self):
        """ Determines two core words in JOS collocation. """
        ppb_components = []
        for c in self.components:
            if c.type != ComponentType.Core:
                continue

            ppb = 4
            for r in c.restrictions:
                ppb = min(r.ppb, ppb)

            ppb_components.append((c, ppb))

        ppb_components = sorted(ppb_components, key=lambda ch: ch[1])
        if len(ppb_components) > 2 and ppb_components[1][1] == ppb_components[2][1]:
            raise RuntimeError("Cannot determine 2 'jedrna polnopomenska beseda' for", self.id)

        for c, _ in ppb_components[:2]:
            c.type = ComponentType.Core2w

    def add_representation(self, n, rep_el, forms):
        """ Adds representation to syntactic structure. """
        assert rep_el.tag == "representation"
        to_add = []
        for el in rep_el:
            assert el.tag == "feature"
            if 'rendition' in el.attrib or 'selection' in el.attrib or 'format' in el.attrib:
                to_add.append(el)
            else:
                logging.warning("Strange representation feature in structure {}. Skipping"
                                .format(self.id))
                continue
        forms[n].append(to_add)

    def match(self, word):
        matches = self.components[0].match(word)
        return [] if matches is None else matches


def build_structures(args):
    """ Builds structures. """
    filename = args['structures']
    no_stats = not args['statistics']

    max_num_components = -1
    with open(filename, 'r', encoding="UTF-8") as fp:
        et = ElementTree.XML(fp.read())

    assert 'system_type' in et.attrib, ValueError('Please update file containing structure descriptions.')
    is_ud = et.attrib['system_type'] == 'UD'
    structures = []
    for structure in et.iter('syntactic_structure'):
        if not ((structure.attrib['type'] == 'collocation' and 'formal_descendants' not in structure.attrib) or structure.attrib['type'] == 'formal'):
            continue
        to_append = SyntacticStructure.from_xml(structure, no_stats, is_ud)
        if to_append is None:
            continue

        structures.append(to_append)
        to_append_len = len(to_append.components) if not to_append.fake_root_included else len(to_append.components) - 1
        max_num_components = max(max_num_components, to_append_len)

    return structures, max_num_components, is_ud



"""
Converter for msd translations and msd conversions to properties
"""
from conversion_utils.jos_msds_and_properties import Converter, Msd, Properties, LEVEL_EXCEPTIONS
from conversion_utils.translate_conllu_jos import get_syn_map

converter = Converter()

syn_map = get_syn_map()


class OptimizedConverter(object):
    def __init__(self):
        self.properties = {category.codes.en.upper(): (category.names.en, [{value.codes.en: (feature.names.en, value.names.en) for value in feature.values} for feature in category.features]) for category in converter.specifications.categories}
        self.lemma_properties = {category.codes.en.upper(): (category.names.en, [{value.codes.en: (feature.names.en, value.names.en) for value in feature.values} for feature in category.features]) for category in converter.specifications.categories}

    def msd_to_properties(self, msd, language, lemma=None, require_valid_flag=False, warn_level_flag=False):
        category_name, category_properties = self.properties[msd[0]]

        properties = {'pos': category_name}
        try:
            for i in range(len(msd) - 1):
                property_char = msd[i + 1]
                if property_char != '-':
                    property_type, property_value = category_properties[i][property_char]
                    properties[property_type] = property_value
            return properties
        except:
            raise 'Msd language might be set incorrectly. Try switching `jos_msd_lang` parameter to `sl`'


optimized_converter = OptimizedConverter()

def msd_to_properties(msd_text, lang, lemma=None):
    """ Converts msd to properties using conversion_utils library. """
    # msd_model = Msd(msd_text, lang)
    return optimized_converter.msd_to_properties(msd_text, lang, lemma=lemma)


def default_msd_to_properties(msd, language, lemma=None, require_valid_flag=False, warn_level_flag=False):
    msd_object = Msd(msd, language)
    return converter.msd_to_properties(msd_object, language, lemma=lemma, require_valid_flag=require_valid_flag, warn_level_flag=warn_level_flag)

def translate_msd(msd_text, lang, lemma=None):
    """ Translates msd using conversion_utils library. """
    if lang == 'en':
        return msd_text
    return converter.properties_to_msd(converter.msd_to_properties(Msd(msd_text, lang), 'en', lemma),
                                       'en').code


def translate_jos_depparse(tag, translate_jos_depparse_to_sl):
    """ Translates jos depparse using conversion_utils library. """
    if not translate_jos_depparse_to_sl:
        return tag
    if tag not in syn_map:
        raise ValueError(f'Tag "{tag}" is not recognized as a valid English tag. You might be using Slovenian depparse system in which case set "jos_depparse_lang" to "sl".')
    return syn_map[tag]



"""
A class for saving statistics.
"""
from collections import defaultdict, Counter

from ast import literal_eval

from cordex.utils.progress_bar import progress
import logging


class WordStats:
    def __init__(self, args, db):
        self.db = db
        self.is_ud = args['is_ud']
        self.all_words = None

        if self.is_ud:
            self.db.init("""CREATE TABLE UniqWords (
                uw_id INTEGER PRIMARY KEY, 
                lemma varchar(64), 
                udpos varchar(32), 
                text varchar(64), 
                frequency int
                )""")
            self.db.init("CREATE TABLE WordCountUPOS (lemma varchar(64), upos varchar(32), frequency int)")
            self.db.init("CREATE INDEX lemma_msd_text_on_uw ON UniqWords (lemma, udpos, text)")
            self.db.init("CREATE INDEX lemma_msd0_on_wc ON WordCountUPOS (lemma, upos)")
        else:
            self.db.init("""CREATE TABLE UniqWords (
                            uw_id INTEGER PRIMARY KEY, 
                            lemma varchar(64), 
                            xpos varchar(16), 
                            text varchar(64), 
                            frequency int
                            )""")
            self.db.init("CREATE TABLE WordCountXPOS (lemma varchar(64), xpos0 char, frequency int)")
            self.db.init("CREATE INDEX lemma_msd_text_on_uw ON UniqWords (lemma, xpos, text)")
            self.db.init("CREATE INDEX lemma_msd0_on_wc ON WordCountXPOS (lemma, xpos0)")
        self.db.init("CREATE TABLE NumWords (id INTEGER PRIMARY KEY, n INTEGER)")

        self.db.init("CREATE INDEX lemma_on_uw ON UniqWords (lemma)")

    def add_words(self, words):
        """ Adds words to database. """
        for w in progress(words, "adding-words"):
            if w.fake_word:
                continue
            if self.is_ud:
                params = {'lemma': w.lemma, 'udpos': str(w.udpos), 'text': w.text}
                res = self.db.execute("""UPDATE UniqWords SET frequency=frequency + 1
                    WHERE lemma=:lemma AND udpos=:udpos AND text=:text""", params)

                if res.rowcount == 0:
                    self.db.execute("""INSERT INTO UniqWords (lemma, udpos, text, frequency) 
                        VALUES (:lemma, :udpos, :text, 1)""", params)
            else:
                params = {'lemma': w.lemma, 'xpos': str(w.xpos), 'text': w.text}
                res = self.db.execute("""UPDATE UniqWords SET frequency=frequency + 1
                                    WHERE lemma=:lemma AND xpos=:xpos AND text=:text""", params)

                if res.rowcount == 0:
                    self.db.execute("""INSERT INTO UniqWords (lemma, xpos, text, frequency) 
                                        VALUES (:lemma, :xpos, :text, 1)""", params)

        self.db.execute("INSERT INTO NumWords (n) VALUES (?)", (len(words),))

    def lowercase_words_under_threshold(self):
        """ Lowercase words that have lowercased version that occur more than 10 % of times in corpus. """
        threshold = 0.1
        all_data = self.db.execute("""SELECT * FROM UniqWords""")
        for i, lemma, pos, text, freq in all_data:
            if text[0].isupper():
                if self.is_ud:
                    lowercased_options = self.db.execute("""SELECT text, frequency FROM UniqWords WHERE lemma=? AND udpos=? AND text=?""", (lemma, pos, text.lower(),))
                else:
                    lowercased_options = self.db.execute(
                        """SELECT text, frequency FROM UniqWords WHERE lemma=? AND xpos=? AND text=?""", (lemma, pos, text.lower(),))

                for lc_text, lc_freq in lowercased_options:
                    if freq >= lc_freq >= freq * threshold:
                        if self.is_ud:
                            self.db.execute(
                                """UPDATE UniqWords SET text=? WHERE lemma=? AND udpos=? AND text=?""",
                                (text.lower(), lemma, pos, text,))
                        else:
                            self.db.execute(
                                """UPDATE UniqWords SET text=? WHERE lemma=? AND xpos=? AND text=?""",
                                (text.lower(), lemma, pos, text,))


    def num_all_words(self):
        """ Counts all words. """

        if self.all_words is None:
            cur = self.db.execute("SELECT sum(n) FROM NumWords")
            self.all_words = int(cur.fetchone()[0])
        return self.all_words

    def generate_renders(self):
        """ Counts frequencies for lemma + msd combinations. """
        step_name = 'generate_renders'
        if self.db.is_step_done(step_name):
            logging.info("Skipping GenerateRenders, already complete")
            return

        lemmas = [lemma for (lemma, ) in self.db.execute("SELECT DISTINCT lemma FROM UniqWords")]
        for lemma in progress(lemmas, 'word-count'):
            if self.is_ud:
                num_words_upos = defaultdict(int)
                for (udpos, freq) in self.db.execute("SELECT udpos, frequency FROM UniqWords WHERE lemma=?", (lemma,)):
                    upos = literal_eval(udpos)['POS']
                    num_words_upos[upos] += freq

                for upos, freq in num_words_upos.items():
                    self.db.execute("INSERT INTO WordCountUPOS (lemma, upos, frequency) VALUES (?,?,?)",
                                    (lemma, upos, freq))

            else:
                num_words_xpos = defaultdict(int)
                for (xpos, freq) in self.db.execute("SELECT xpos, frequency FROM UniqWords WHERE lemma=?", (lemma,)):
                    xpos0 = xpos[0]
                    num_words_xpos[xpos0] += freq

                for xpos0, freq in num_words_xpos.items():
                    self.db.execute("INSERT INTO WordCountXPOS (lemma, xpos0, frequency) VALUES (?,?,?)",
                                    (lemma, xpos0, freq))

        self.db.step_is_done(step_name)

    def render(self, lemma, msd):
        """ Returns most frequent word for specific lemma+msd pair. """
        if self.is_ud:
            statement = """SELECT text FROM UniqWords WHERE 
                           lemma=:lemma AND udpos=:udpos ORDER BY frequency DESC"""

            cur = self.db.execute(statement, {"lemma": lemma, "udpos": msd})
        else:
            statement = """SELECT text FROM UniqWords WHERE 
                        lemma=:lemma AND xpos=:xpos ORDER BY frequency DESC"""

            cur = self.db.execute(statement, {"lemma": lemma, "xpos": msd})
        fetch = cur.fetchone()
        if fetch is None:
            return None

        return fetch[0]

    def available_words(self, lemma):
        """ Lists possible words for agreements and lists them in descending order. """
        if self.is_ud:
            statement = """SELECT udpos, text, frequency FROM UniqWords WHERE 
                        lemma=:lemma ORDER BY frequency DESC"""
            for udpos, text, _f in self.db.execute(statement, {'lemma': lemma}):
                yield literal_eval(udpos), text, lemma
        else:
            statement = """SELECT xpos, text, frequency FROM UniqWords WHERE 
                        lemma=:lemma ORDER BY frequency DESC"""
            for xpos, text, _f in self.db.execute(statement, {'lemma': lemma}):
                yield xpos, text, lemma

    def num_words(self, lemma, msd0):
        """ Returns first word frequency when lemma and msd match. """
        if self.is_ud:
            statement = "SELECT frequency FROM WordCountUPOS WHERE lemma=? AND upos=? LIMIT 1"
        else:
            statement = "SELECT frequency FROM WordCountXPOS WHERE lemma=? AND xpos0=? LIMIT 1"
        cur = self.db.execute(statement, (lemma, msd0))
        result = cur.fetchone()[0]
        return result



import filecmp
import os
import shutil

import pytest
import cordex
from tests import *
from tests.correct_output import OUTPUT_TOKEN_OUTPUT, OUTPUT_GET_LIST


@pytest.fixture(scope="module")
def clear_output():
    """ Erases temporary files and creates new folders. """
    output_mapper_dir = os.path.join(OUTPUT_DIR, "tmp_collocation_sentence_mapper")
    if os.path.exists(output_mapper_dir):
        shutil.rmtree(output_mapper_dir)
    os.makedirs(output_mapper_dir)
    output_dir = os.path.join(OUTPUT_DIR, "izhod.csv")

    return os.path.join(output_mapper_dir, 'mapper.csv'), output_dir


def compare_directories(cor_output, output):
    for c_o, o in zip(list(os.walk(cor_output)), list(os.walk(output))):
        c_o_subdir, c_o_dirs, c_o_files = c_o
        o_subdir, o_dirs, o_files = o
        for c_o_f, o_f in zip(c_o_files, o_files):
            assert filecmp.cmp(os.path.join(c_o_subdir, c_o_f), os.path.join(o_subdir, o_f))


def test_download(clear_output):
    """ Test for cordex download. """
    cordex.download()


def test_tei_multiple_ud_documents(clear_output):
    """ Test for tei multiple documents. """
    output_mapper_dir, output_dir = clear_output

    extractor = cordex.Pipeline(os.path.join(STRUCTURES_DIR, "structures_UD.xml"),
                           collocation_sentence_map_dest=output_mapper_dir,
                           lang='en', jos_msd_lang='en')
    extraction = extractor(os.path.join(INPUT_DIR, "gigafida_example_tei_small"))
    extraction.write(output_dir, separator=',')
    compare_directories(os.path.join(CORRECT_OUTPUT_DIR, 'output_tei_multiple_ud_documents'), os.path.join(OUTPUT_DIR))


def test_tei_single_jos_document(clear_output):
    """ Test for tei single document. """
    output_mapper_dir, output_dir = clear_output

    extractor = cordex.Pipeline(os.path.join(STRUCTURES_DIR, "structures_JOS.xml"),
                           collocation_sentence_map_dest=output_mapper_dir, jos_msd_lang='sl')
    extraction = extractor(os.path.join(INPUT_DIR, "ssj500k.small.xml"))
    extraction.write(output_dir, separator=',')
    compare_directories(os.path.join(CORRECT_OUTPUT_DIR, 'output_tei_single_jos_document'), os.path.join(OUTPUT_DIR))


def test_tei_single_ud_document(clear_output):
    """ Test for tei single document. Also tests no statistics output"""
    output_mapper_dir, output_dir = clear_output

    extractor = cordex.Pipeline(os.path.join(STRUCTURES_DIR, "structures_UD.xml"),
                           collocation_sentence_map_dest=output_mapper_dir, statistics=False)
    extraction = extractor(os.path.join(INPUT_DIR, "ssj500k.small.xml"))
    extraction.write(output_dir, separator=',')
    compare_directories(os.path.join(CORRECT_OUTPUT_DIR, 'output_tei_single_ud_document'), os.path.join(OUTPUT_DIR))


def test_conllu_single_jos_document(clear_output):
    """ Test for conllu single jos document. """
    output_mapper_dir, output_dir = clear_output

    extractor = cordex.Pipeline(os.path.join(STRUCTURES_DIR, "structures_JOS.xml"),
                           collocation_sentence_map_dest=output_mapper_dir, jos_depparse_lang="en")
    extraction = extractor(os.path.join(INPUT_DIR, "test_conllu_jos_small.conllu"))
    extraction.write(output_dir, separator=',')

    compare_directories(os.path.join(CORRECT_OUTPUT_DIR, 'output_conllu_single_jos_document'), os.path.join(OUTPUT_DIR))


def test_conllu_multiple_ud_documents(clear_output):
    """ Test for conllu multiple ud documents. """
    output_mapper_dir, output_dir = clear_output

    extractor = cordex.Pipeline(os.path.join(STRUCTURES_DIR, "structures_UD.xml"),
                           collocation_sentence_map_dest=output_mapper_dir)
    extraction = extractor(os.path.join(INPUT_DIR, "gigafida_example_conllu_small"))
    extraction.write(output_dir, separator=',')

    compare_directories(os.path.join(CORRECT_OUTPUT_DIR, 'output_conllu_multiple_ud_documents'), os.path.join(OUTPUT_DIR))


def test_conllu_single_ud_document(clear_output):
    """ Test for conllu single ud document. """
    output_mapper_dir, output_dir = clear_output

    extractor = cordex.Pipeline(os.path.join(STRUCTURES_DIR, "structures_UD.xml"),
                           collocation_sentence_map_dest=output_mapper_dir)
    extraction = extractor(os.path.join(INPUT_DIR, "ssj500k.small.conllu"))
    extraction.write(output_dir, separator=',')

    compare_directories(os.path.join(CORRECT_OUTPUT_DIR, 'output_conllu_single_ud_document'), os.path.join(OUTPUT_DIR))


def test_get_list():
    """ Test for conllu single ud document. """
    extractor = cordex.Pipeline(os.path.join(STRUCTURES_DIR, "structures_UD.xml"))
    extraction = extractor(os.path.join(INPUT_DIR, "ssj500k.small.conllu"))
    output = extraction.get_list()

    assert OUTPUT_GET_LIST == str(output)


def test_tei_multiple_jos_documents(clear_output):
    """ Test for tei multiple documents. """
    output_mapper_dir, output_dir = clear_output

    extractor = cordex.Pipeline(os.path.join(STRUCTURES_DIR, "structures_JOS.xml"),
                           collocation_sentence_map_dest=output_mapper_dir,
                           lang='en', jos_msd_lang='sl')
    extraction = extractor(os.path.join(INPUT_DIR, "gigafida_example_tei_small"))
    extraction.write(output_dir, separator=',')
    compare_directories(os.path.join(CORRECT_OUTPUT_DIR, 'output_tei_multiple_jos_documents'), os.path.join(OUTPUT_DIR))


def test_lookup_api(clear_output):
    """ Test for conllu single jos document. """
    output_mapper_dir, output_dir = clear_output

    extractor = cordex.Pipeline(os.path.join(STRUCTURES_DIR, "structures_JOS.xml"),
                           collocation_sentence_map_dest=output_mapper_dir, lookup_api=True, jos_depparse_lang="en")
    extraction = extractor(os.path.join(INPUT_DIR, "test_conllu_jos_small.conllu"))
    extraction.write(output_dir, separator=',')

    compare_directories(os.path.join(CORRECT_OUTPUT_DIR, 'output_conllu_single_jos_document'), os.path.join(OUTPUT_DIR))


def test_no_lookup_jos(clear_output):
    """ Test for lookup api. """
    output_mapper_dir, output_dir = clear_output

    extractor = cordex.Pipeline(os.path.join(STRUCTURES_DIR, "structures_JOS.xml"),
                           collocation_sentence_map_dest=output_mapper_dir, lookup_lexicon=None, jos_msd_lang='sl')
    extraction = extractor(os.path.join(INPUT_DIR, "gigafida_example_tei_small"))
    extraction.write(output_dir, separator=',')
    compare_directories(os.path.join(CORRECT_OUTPUT_DIR, 'output_no_lookup'), os.path.join(OUTPUT_DIR))



"""
Connections to lexicon via API or file reading
"""
import logging
import lzma
import pickle
import requests
from cordex.utils.codes_tagset import TAGSET, CODES
from cordex.utils.converter import msd_to_properties, default_msd_to_properties
from cordex.utils.progress_bar import progress

SUPPORTED_LOOKUP_LANGUAGES = ['sl']

class LookupApi:
    """ A class for managing API calls. """
    def __init__(self, base_url):
        self.base_url = base_url
        self.forms_data = {}
        self.executed = False
        self.search_lexeme_batch_size = 7500
        self.retrieve_lexeme_batch_size = 500

    def call_api(self, url, json, i, end_i):
        """ Calls api for json[i:end_i]. If response is anything but status code 200, make two calls of halved batch size. """
        response = requests.post(url, json=json[i:end_i])
        if response.status_code != 200:
            new_batch_size = int((end_i - i)/2)
            if new_batch_size < 50:
                raise Exception('Unable to obtain data from API!')
            logging.info(f'{url} with batch size {end_i - i} returned error! Processing with batch size {new_batch_size}')
            response_json = []
            response_json.extend(self.call_api(url, json, i, i + new_batch_size))
            response_json.extend(self.call_api(url, json, i + new_batch_size, end_i))
        else:
            response_json = response.json()
        return response_json
    def search_lexeme(self, post_request):
        """ Executes search-batch request for /search-batch/lexeme/ """
        all_lexemes = []
        batch_size = self.search_lexeme_batch_size
        ran = range(0, len(post_request), batch_size)
        for i in progress(ran,
                          "search lexeme requests", total=len(ran)):
            end_i = min(i + batch_size, len(post_request))

            response_json = self.call_api(self.base_url + '/search-batch/lexeme/', post_request, i, end_i)
            all_lexemes.extend(response_json)
        return all_lexemes

    def retrieve_lexeme(self, post_request):
        """ Executes retrieve-batch request for /retrieve-batch/lexeme/ """
        all_lexemes = []
        batch_size = self.retrieve_lexeme_batch_size
        ran = range(0, len(post_request), batch_size)
        for i in progress(ran,
                          "retrieve lexeme requests", total=len(ran)):
            end_i = min(i + batch_size, len(post_request))
            response_json = self.call_api(self.base_url + f'/retrieve-batch/lexeme/', post_request, i, end_i)
            all_lexemes.extend(response_json)
        return all_lexemes

    def find_lemma(self, lemma_key):
        """ Finds most frequent lemma and its msd based on lemma key. """
        lemma = self.forms_data[lemma_key]['lemma']
        for form in self.forms_data[lemma_key]['forms']:
            if form[0] == lemma:
                return form
        return None

    def find_form(self, lemma_key, form_restrictions):
        """ Finds most frequent proper form based on lemma key and form restrictions. """
        for form in self.forms_data[lemma_key]['forms']:
            match = True
            for k, v in form_restrictions.items():
                k_lowercased = k.lower()
                if k_lowercased not in form[2] or form[2][k_lowercased] != v:
                    match = False
                    break

            if match:
                return form
        return None

    def execute_requests(self, num_representations, db, all_representations):
        """ Creates and executes queries on API from fake representations. """
        post_request = []
        i = 0
        # get data for lexeme post
        for cid, sid in progress(db.execute("SELECT collocation_id, structure_id FROM Collocations"),
                                 "representations-gather-lexeme", total=num_representations):

            representations = all_representations[i]
            for cid, reps in representations.items():
                for rep in reps:
                    if rep.processed_api:
                        for req in rep.api_request:
                            if req['type'] == 'lemma':
                                post_request.append((f"{req['lemma']}_{req['category']}_{'_'.join([f'{k},{v}' for k, v in req['lemma_features'].items()])}", {
                                    'lemma': req['lemma'],
                                    "category": req['category'],
                                    'features': req['lemma_features']
                                }, {'lemma': req['lemma'], 'key': 'lemma'}))
                            elif req['type'] == 'msd' or req['type'] == 'agreement':
                                post_request.append((f"{req['lemma']}_{req['category']}_{'_'.join([f'{k},{v}' for k, v in req['lemma_features'].items()])}", {
                                    'lemma': req['lemma'],
                                    "category": req['category'],
                                    'features': req['lemma_features']
                                }, {'lemma': req['lemma'], 'key': f"{'_'.join([f'{k},{v}' for k, v in req['form_features'].items()])}", 'form_feature': req['form_features']}))
                            else:
                                raise f'req["type"] - {req["type"]} should not exist.'
            i += 1

        # deduplicate post_request
        search_post_request = []
        forms_data = {}
        unique_post_requests = set()
        for el in post_request:
            if el[0] not in unique_post_requests:
                unique_post_requests.add(el[0])
                search_post_request.append(el[1])
                forms_data[el[0]] = {'forms': [], 'lemma': el[1]['lemma']}

        lexeme_responses = self.search_lexeme(search_post_request)

        for form_data_key, lexeme_response in zip(forms_data.keys(), lexeme_responses):
            forms_data[form_data_key]['lexeme_ids'] = [r['id'] for r in lexeme_response['data']]

        retrieve_post_request = [{
                            'lexeme_id': data_el['id'],
                            "msd-language": "en",
                            "extra-data": [ "forms-orthography" ],
                            'corpus_id': 2
                        } for el in lexeme_responses for data_el in el['data']]

        forms_response = self.retrieve_lexeme(retrieve_post_request)

        form_responses_linked = {post_text['lexeme_id']: forms['data'] for post_text, forms in zip(retrieve_post_request, forms_response)}

        # link words and msds with keys
        for lemma_key, lemma_content in forms_data.items():
            complacent_forms = []
            for l_id in lemma_content['lexeme_ids']:
                complacent_forms.extend(form_responses_linked[l_id]['forms'])
            possible_forms = []
            for form in complacent_forms:
                form_features = default_msd_to_properties(form['msd'], 'en', lemma=lemma_content['lemma']).form_feature_map
                for subform in form['forms']:
                    for subsubform in subform:
                        frequency = subsubform['frequency'] if 'frequency' in subsubform and subsubform[
                            'frequency'] else 0
                        possible_forms.append((subsubform['text'], frequency, form_features, form['msd']))

            possible_forms = sorted(possible_forms, key=lambda x: -x[1])
            lemma_content['forms'] = possible_forms

        self.forms_data = forms_data
        self.executed = True

        return forms_data

class LookupLexicon:
    """ Object that access lookup lexicon file. """
    def __init__(self, file_path=''):
        self.file_data = None
        self.init_file_reading(file_path)

    def init_file_reading(self, file_path):
        """ Loads lookup lexicon file into memory. """
        with lzma.open(file_path, "rb") as f:
            self.file_data = pickle.load(f)

    def get_word_form(self, lemma, msd, data, align_msd=False, find_lemma_msd=False, align_lemma=None):
        """ Returns word form from lemma, msd restrictions and/or alignment msd. """
        # handles cases when we are searching msds of lemmas
        if lemma in self.file_data and find_lemma_msd:
            for (word_form_features, form_representations, xpos, form_frequency) in self.file_data[lemma]:
                if lemma == form_representations:
                    return xpos, lemma, form_representations

        # modify msd as required
        form_features = {}
        if 'msd' in data:
            form_features = {k.lower(): v.lower() for k, v in data['msd'].items()}

        if align_msd and 'agreement' in data:
            # get agreement feature properties
            agreement_properties = default_msd_to_properties(align_msd, 'en', align_lemma)

            missing_form_features = False
            for agreement_name in data['agreement']:
                if agreement_name in agreement_properties.form_feature_map:
                    form_features[agreement_name] = agreement_properties.form_feature_map[agreement_name]
                elif agreement_name in agreement_properties.lexeme_feature_map:
                    form_features[agreement_name] = agreement_properties.lexeme_feature_map[agreement_name]

                if agreement_name not in agreement_properties.form_feature_map and agreement_name not in agreement_properties.lexeme_feature_map:
                    missing_form_features = True
                    break

            if missing_form_features:
                return None, None, None

        if lemma in self.file_data:
            for (word_form_features, form_representations, xpos, form_frequency) in self.file_data[lemma]:
                fits = True
                for d_m_attribute, d_m_value in form_features.items():
                    if d_m_attribute not in word_form_features or d_m_value != word_form_features[d_m_attribute]:
                        fits = False
                        break
                if fits:
                    break
            return xpos, lemma, form_representations
        return None, None, None



"""
Translation related functions.
"""


def translate_sl_en(xpos):
    return xpos


MSD_TRANSLATE = {
    "Somei": "Ncmsn",
    "Somer": "Ncmsg",
    "Somed": "Ncmsd",
    "Sometn": "Ncmsan",
    "Sometd": "Ncmsay",
    "Somem": "Ncmsl",
    "Someo": "Ncmsi",
    "Somdi": "Ncmdn",
    "Somdr": "Ncmdg",
    "Somdd": "Ncmdd",
    "Somdt": "Ncmda",
    "Somdm": "Ncmdl",
    "Somdo": "Ncmdi",
    "Sommi": "Ncmpn",
    "Sommr": "Ncmpg",
    "Sommd": "Ncmpd",
    "Sommt": "Ncmpa",
    "Sommm": "Ncmpl",
    "Sommo": "Ncmpi",
    "Sozei": "Ncfsn",
    "Sozer": "Ncfsg",
    "Sozed": "Ncfsd",
    "Sozet": "Ncfsa",
    "Sozem": "Ncfsl",
    "Sozeo": "Ncfsi",
    "Sozdi": "Ncfdn",
    "Sozdr": "Ncfdg",
    "Sozdd": "Ncfdd",
    "Sozdt": "Ncfda",
    "Sozdm": "Ncfdl",
    "Sozdo": "Ncfdi",
    "Sozmi": "Ncfpn",
    "Sozmr": "Ncfpg",
    "Sozmd": "Ncfpd",
    "Sozmt": "Ncfpa",
    "Sozmm": "Ncfpl",
    "Sozmo": "Ncfpi",
    "Sosei": "Ncnsn",
    "Soser": "Ncnsg",
    "Sosed": "Ncnsd",
    "Soset": "Ncnsa",
    "Sosem": "Ncnsl",
    "Soseo": "Ncnsi",
    "Sosdi": "Ncndn",
    "Sosdr": "Ncndg",
    "Sosdd": "Ncndd",
    "Sosdt": "Ncnda",
    "Sosdm": "Ncndl",
    "Sosdo": "Ncndi",
    "Sosmi": "Ncnpn",
    "Sosmr": "Ncnpg",
    "Sosmd": "Ncnpd",
    "Sosmt": "Ncnpa",
    "Sosmm": "Ncnpl",
    "Sosmo": "Ncnpi",
    "Slmei": "Npmsn",
    "Slmer": "Npmsg",
    "Slmed": "Npmsd",
    "Slmetn": "Npmsan",
    "Slmetd": "Npmsay",
    "Slmem": "Npmsl",
    "Slmeo": "Npmsi",
    "Slmdi": "Npmdn",
    "Slmdr": "Npmdg",
    "Slmdd": "Npmdd",
    "Slmdt": "Npmda",
    "Slmdm": "Npmdl",
    "Slmdo": "Npmdi",
    "Slmmi": "Npmpn",
    "Slmmr": "Npmpg",
    "Slmmd": "Npmpd",
    "Slmmt": "Npmpa",
    "Slmmm": "Npmpl",
    "Slmmo": "Npmpi",
    "Slzei": "Npfsn",
    "Slzer": "Npfsg",
    "Slzed": "Npfsd",
    "Slzet": "Npfsa",
    "Slzem": "Npfsl",
    "Slzeo": "Npfsi",
    "Slzdi": "Npfdn",
    "Slzdr": "Npfdg",
    "Slzdd": "Npfdd",
    "Slzdt": "Npfda",
    "Slzdm": "Npfdl",
    "Slzdo": "Npfdi",
    "Slzmi": "Npfpn",
    "Slzmr": "Npfpg",
    "Slzmd": "Npfpd",
    "Slzmt": "Npfpa",
    "Slzmm": "Npfpl",
    "Slzmo": "Npfpi",
    "Slsei": "Npnsn",
    "Slser": "Npnsg",
    "Slsed": "Npnsd",
    "Slset": "Npnsa",
    "Slsem": "Npnsl",
    "Slseo": "Npnsi",
    "Slsmi": "Npnpn",
    "Slsmr": "Npnpg",
    "Slsmd": "Npnpd",
    "Slsmt": "Npnpa",
    "Slsmm": "Npnpl",
    "Slsmo": "Npnpi",
    "Ggdn": "Vmen",
    "Ggdm": "Vmeu",
    "Ggdd-em": "Vmep-sm",
    "Ggdd-ez": "Vmep-sf",
    "Ggdd-es": "Vmep-sn",
    "Ggdd-mm": "Vmep-pm",
    "Ggdd-mz": "Vmep-pf",
    "Ggdd-ms": "Vmep-pn",
    "Ggdd-dm": "Vmep-dm",
    "Ggdd-dz": "Vmep-df",
    "Ggdd-ds": "Vmep-dn",
    "Ggdspe": "Vmer1s",
    "Ggdspm": "Vmer1p",
    "Ggdspd": "Vmer1d",
    "Ggdsde": "Vmer2s",
    "Ggdsdm": "Vmer2p",
    "Ggdsdd": "Vmer2d",
    "Ggdste": "Vmer3s",
    "Ggdstm": "Vmer3p",
    "Ggdstd": "Vmer3d",
    "Ggdvpm": "Vmem1p",
    "Ggdvpd": "Vmem1d",
    "Ggdvde": "Vmem2s",
    "Ggdvdm": "Vmem2p",
    "Ggdvdd": "Vmem2d",
    "Ggnn": "Vmpn",
    "Ggnm": "Vmpu",
    "Ggnd-em": "Vmpp-sm",
    "Ggnd-ez": "Vmpp-sf",
    "Ggnd-es": "Vmpp-sn",
    "Ggnd-mm": "Vmpp-pm",
    "Ggnd-mz": "Vmpp-pf",
    "Ggnd-ms": "Vmpp-pn",
    "Ggnd-dm": "Vmpp-dm",
    "Ggnd-dz": "Vmpp-df",
    "Ggnd-ds": "Vmpp-dn",
    "Ggnspe": "Vmpr1s",
    "Ggnspe-n": "Vmpr1s-n",
    "Ggnspe-d": "Vmpr1s-y",
    "Ggnspm": "Vmpr1p",
    "Ggnspm-n": "Vmpr1p-n",
    "Ggnspm-d": "Vmpr1p-y",
    "Ggnspd": "Vmpr1d",
    "Ggnspd-n": "Vmpr1d-n",
    "Ggnspd-d": "Vmpr1d-y",
    "Ggnsde": "Vmpr2s",
    "Ggnsde-n": "Vmpr2s-n",
    "Ggnsde-d": "Vmpr2s-y",
    "Ggnsdm": "Vmpr2p",
    "Ggnsdm-n": "Vmpr2p-n",
    "Ggnsdm-d": "Vmpr2p-y",
    "Ggnsdd": "Vmpr2d",
    "Ggnsdd-n": "Vmpr2d-n",
    "Ggnsdd-d": "Vmpr2d-y",
    "Ggnste": "Vmpr3s",
    "Ggnste-n": "Vmpr3s-n",
    "Ggnste-d": "Vmpr3s-y",
    "Ggnstm": "Vmpr3p",
    "Ggnstm-n": "Vmpr3p-n",
    "Ggnstm-d": "Vmpr3p-y",
    "Ggnstd": "Vmpr3d",
    "Ggnstd-n": "Vmpr3d-n",
    "Ggnstd-d": "Vmpr3d-y",
    "Ggnvpm": "Vmpm1p",
    "Ggnvpd": "Vmpm1d",
    "Ggnvde": "Vmpm2s",
    "Ggnvdm": "Vmpm2p",
    "Ggnvdd": "Vmpm2d",
    "Ggvn": "Vmbn",
    "Ggvm": "Vmbu",
    "Ggvd-em": "Vmbp-sm",
    "Ggvd-ez": "Vmbp-sf",
    "Ggvd-es": "Vmbp-sn",
    "Ggvd-mm": "Vmbp-pm",
    "Ggvd-mz": "Vmbp-pf",
    "Ggvd-ms": "Vmbp-pn",
    "Ggvd-dm": "Vmbp-dm",
    "Ggvd-dz": "Vmbp-df",
    "Ggvd-ds": "Vmbp-dn",
    "Ggvspe": "Vmbr1s",
    "Ggvspm": "Vmbr1p",
    "Ggvspd": "Vmbr1d",
    "Ggvspdz": "Vmbr1df",
    "Ggvsde": "Vmbr2s",
    "Ggvsdm": "Vmbr2p",
    "Ggvsdd": "Vmbr2d",
    "Ggvste": "Vmbr3s",
    "Ggvstm": "Vmbr3p",
    "Ggvstd": "Vmbr3d",
    "Ggvppe": "Vmbf1s",
    "Ggvppm": "Vmbf1p",
    "Ggvppd": "Vmbf1d",
    "Ggvpde": "Vmbf2s",
    "Ggvpdm": "Vmbf2p",
    "Ggvpdd": "Vmbf2d",
    "Ggvpte": "Vmbf3s",
    "Ggvptm": "Vmbf3p",
    "Ggvptd": "Vmbf3d",
    "Ggvvpm": "Vmbm1p",
    "Ggvvpd": "Vmbm1d",
    "Ggvvde": "Vmbm2s",
    "Ggvvdm": "Vmbm2p",
    "Ggvvdd": "Vmbm2d",
    "Gp-n": "Va-n",
    "Gp-m": "Va-u",
    "Gp-d-em": "Va-p-sm",
    "Gp-d-ez": "Va-p-sf",
    "Gp-d-es": "Va-p-sn",
    "Gp-d-mm": "Va-p-pm",
    "Gp-d-mz": "Va-p-pf",
    "Gp-d-ms": "Va-p-pn",
    "Gp-d-dm": "Va-p-dm",
    "Gp-d-dz": "Va-p-df",
    "Gp-d-ds": "Va-p-dn",
    "Gp-spe-n": "Va-r1s-n",
    "Gp-spe-d": "Va-r1s-y",
    "Gp-spm-n": "Va-r1p-n",
    "Gp-spm-d": "Va-r1p-y",
    "Gp-spd-n": "Va-r1d-n",
    "Gp-spd-d": "Va-r1d-y",
    "Gp-spdzn": "Va-r1dfn",
    "Gp-spmzd": "Va-r1pfy",
    "Gp-sde-n": "Va-r2s-n",
    "Gp-sde-d": "Va-r2s-y",
    "Gp-sdm-n": "Va-r2p-n",
    "Gp-sdm-d": "Va-r2p-y",
    "Gp-sdd-n": "Va-r2d-n",
    "Gp-sdd-d": "Va-r2d-y",
    "Gp-ste-n": "Va-r3s-n",
    "Gp-ste-d": "Va-r3s-y",
    "Gp-stm-n": "Va-r3p-n",
    "Gp-stm-d": "Va-r3p-y",
    "Gp-std-n": "Va-r3d-n",
    "Gp-std-d": "Va-r3d-y",
    "Gp-ppe-n": "Va-f1s-n",
    "Gp-ppe-d": "Va-f1s-y",
    "Gp-ppm-n": "Va-f1p-n",
    "Gp-ppm-d": "Va-f1p-y",
    "Gp-ppmzn": "Va-f1pfy",
    "Gp-ppd": "Va-f1d",
    "Gp-ppdzn": "Va-f1dfn",
    "Gp-ppdzd": "Va-f1dfy",
    "Gp-ppd-n": "Va-f1d-n",
    "Gp-ppd-d": "Va-f1d-y",
    "Gp-pde-n": "Va-f2s-n",
    "Gp-pde-d": "Va-f2s-y",
    "Gp-pdm-n": "Va-f2p-n",
    "Gp-pdm-d": "Va-f2p-y",
    "Gp-pdd-n": "Va-f2d-n",
    "Gp-pte-n": "Va-f3s-n",
    "Gp-pte-d": "Va-f3s-y",
    "Gp-ptm-n": "Va-f3p-n",
    "Gp-ptm-d": "Va-f3p-y",
    "Gp-ptd-n": "Va-f3d-n",
    "Gp-ptd-d": "Va-f3d-y",
    "Gp-g": "Va-c",
    "Gp-g---d": "Va-c---y",
    "Gp-vpm": "Va-m1p",
    "Gp-vpd": "Va-m1d",
    "Gp-vde": "Va-m2s",
    "Gp-vdm": "Va-m2p",
    "Gp-vdd": "Va-m2d",
    "Ppnmein": "Agpmsnn",
    "Ppnmeid": "Agpmsny",
    "Ppnmer": "Agpmsg",
    "Ppnmed": "Agpmsd",
    "Ppnmet": "Agpmsa",
    "Ppnmetn": "Agpmsan",
    "Ppnmetd": "Agpmsay",
    "Ppnmem": "Agpmsl",
    "Ppnmeo": "Agpmsi",
    "Ppnmdi": "Agpmdn",
    "Ppnmdr": "Agpmdg",
    "Ppnmdd": "Agpmdd",
    "Ppnmdt": "Agpmda",
    "Ppnmdm": "Agpmdl",
    "Ppnmdo": "Agpmdi",
    "Ppnmmi": "Agpmpn",
    "Ppnmmr": "Agpmpg",
    "Ppnmmd": "Agpmpd",
    "Ppnmmt": "Agpmpa",
    "Ppnmmm": "Agpmpl",
    "Ppnmmo": "Agpmpi",
    "Ppnzei": "Agpfsn",
    "Ppnzer": "Agpfsg",
    "Ppnzed": "Agpfsd",
    "Ppnzet": "Agpfsa",
    "Ppnzem": "Agpfsl",
    "Ppnzeo": "Agpfsi",
    "Ppnzdi": "Agpfdn",
    "Ppnzdr": "Agpfdg",
    "Ppnzdd": "Agpfdd",
    "Ppnzdt": "Agpfda",
    "Ppnzdm": "Agpfdl",
    "Ppnzdo": "Agpfdi",
    "Ppnzmi": "Agpfpn",
    "Ppnzmr": "Agpfpg",
    "Ppnzmd": "Agpfpd",
    "Ppnzmt": "Agpfpa",
    "Ppnzmm": "Agpfpl",
    "Ppnzmo": "Agpfpi",
    "Ppnsei": "Agpnsn",
    "Ppnser": "Agpnsg",
    "Ppnsed": "Agpnsd",
    "Ppnset": "Agpnsa",
    "Ppnsem": "Agpnsl",
    "Ppnseo": "Agpnsi",
    "Ppnsdi": "Agpndn",
    "Ppnsdr": "Agpndg",
    "Ppnsdd": "Agpndd",
    "Ppnsdt": "Agpnda",
    "Ppnsdm": "Agpndl",
    "Ppnsdo": "Agpndi",
    "Ppnsmi": "Agpnpn",
    "Ppnsmr": "Agpnpg",
    "Ppnsmd": "Agpnpd",
    "Ppnsmt": "Agpnpa",
    "Ppnsmm": "Agpnpl",
    "Ppnsmo": "Agpnpi",
    "Pppmeid": "Agcmsny",
    "Pppmer": "Agcmsg",
    "Pppmed": "Agcmsd",
    "Pppmet": "Agcmsa",
    "Pppmetd": "Agcmsay",
    "Pppmem": "Agcmsl",
    "Pppmeo": "Agcmsi",
    "Pppmdi": "Agcmdn",
    "Pppmdr": "Agcmdg",
    "Pppmdd": "Agcmdd",
    "Pppmdt": "Agcmda",
    "Pppmdm": "Agcmdl",
    "Pppmdo": "Agcmdi",
    "Pppmmi": "Agcmpn",
    "Pppmmr": "Agcmpg",
    "Pppmmd": "Agcmpd",
    "Pppmmt": "Agcmpa",
    "Pppmmm": "Agcmpl",
    "Pppmmo": "Agcmpi",
    "Pppzei": "Agcfsn",
    "Pppzer": "Agcfsg",
    "Pppzed": "Agcfsd",
    "Pppzet": "Agcfsa",
    "Pppzem": "Agcfsl",
    "Pppzeo": "Agcfsi",
    "Pppzdi": "Agcfdn",
    "Pppzdr": "Agcfdg",
    "Pppzdd": "Agcfdd",
    "Pppzdt": "Agcfda",
    "Pppzdm": "Agcfdl",
    "Pppzdo": "Agcfdi",
    "Pppzmi": "Agcfpn",
    "Pppzmr": "Agcfpg",
    "Pppzmd": "Agcfpd",
    "Pppzmt": "Agcfpa",
    "Pppzmm": "Agcfpl",
    "Pppzmo": "Agcfpi",
    "Pppsei": "Agcnsn",
    "Pppser": "Agcnsg",
    "Pppsed": "Agcnsd",
    "Pppset": "Agcnsa",
    "Pppsem": "Agcnsl",
    "Pppseo": "Agcnsi",
    "Pppsdi": "Agcndn",
    "Pppsdr": "Agcndg",
    "Pppsdd": "Agcndd",
    "Pppsdt": "Agcnda",
    "Pppsdm": "Agcndl",
    "Pppsdo": "Agcndi",
    "Pppsmi": "Agcnpn",
    "Pppsmr": "Agcnpg",
    "Pppsmd": "Agcnpd",
    "Pppsmt": "Agcnpa",
    "Pppsmm": "Agcnpl",
    "Pppsmo": "Agcnpi",
    "Ppsmeid": "Agsmsny",
    "Ppsmer": "Agsmsg",
    "Ppsmed": "Agsmsd",
    "Ppsmet": "Agsmsa",
    "Ppsmetd": "Agsmsay",
    "Ppsmem": "Agsmsl",
    "Ppsmeo": "Agsmsi",
    "Ppsmdi": "Agsmdn",
    "Ppsmdr": "Agsmdg",
    "Ppsmdd": "Agsmdd",
    "Ppsmdt": "Agsmda",
    "Ppsmdm": "Agsmdl",
    "Ppsmdo": "Agsmdi",
    "Ppsmmi": "Agsmpn",
    "Ppsmmr": "Agsmpg",
    "Ppsmmd": "Agsmpd",
    "Ppsmmt": "Agsmpa",
    "Ppsmmm": "Agsmpl",
    "Ppsmmo": "Agsmpi",
    "Ppszei": "Agsfsn",
    "Ppszer": "Agsfsg",
    "Ppszed": "Agsfsd",
    "Ppszet": "Agsfsa",
    "Ppszem": "Agsfsl",
    "Ppszeo": "Agsfsi",
    "Ppszdi": "Agsfdn",
    "Ppszdr": "Agsfdg",
    "Ppszdd": "Agsfdd",
    "Ppszdt": "Agsfda",
    "Ppszdm": "Agsfdl",
    "Ppszdo": "Agsfdi",
    "Ppszmi": "Agsfpn",
    "Ppszmr": "Agsfpg",
    "Ppszmd": "Agsfpd",
    "Ppszmt": "Agsfpa",
    "Ppszmm": "Agsfpl",
    "Ppszmo": "Agsfpi",
    "Ppssei": "Agsnsn",
    "Ppsser": "Agsnsg",
    "Ppssed": "Agsnsd",
    "Ppsset": "Agsnsa",
    "Ppssem": "Agsnsl",
    "Ppsseo": "Agsnsi",
    "Ppssdi": "Agsndn",
    "Ppssdr": "Agsndg",
    "Ppssdd": "Agsndd",
    "Ppssdt": "Agsnda",
    "Ppssdm": "Agsndl",
    "Ppssdo": "Agsndi",
    "Ppssmi": "Agsnpn",
    "Ppssmr": "Agsnpg",
    "Ppssmd": "Agsnpd",
    "Ppssmt": "Agsnpa",
    "Ppssmm": "Agsnpl",
    "Ppssmo": "Agsnpi",
    "Psnmein": "Aspmsnn",
    "Psnmer": "Aspmsg",
    "Psnmed": "Aspmsd",
    "Psnmet": "Aspmsa",
    "Psnmetn": "Aspmsan",
    "Psnmem": "Aspmsl",
    "Psnmeo": "Aspmsi",
    "Psnmdi": "Aspmdn",
    "Psnmdr": "Aspmdg",
    "Psnmdd": "Aspmdd",
    "Psnmdt": "Aspmda",
    "Psnmdm": "Aspmdl",
    "Psnmdo": "Aspmdi",
    "Psnmmi": "Aspmpn",
    "Psnmmr": "Aspmpg",
    "Psnmmd": "Aspmpd",
    "Psnmmt": "Aspmpa",
    "Psnmmm": "Aspmpl",
    "Psnmmo": "Aspmpi",
    "Psnzei": "Aspfsn",
    "Psnzer": "Aspfsg",
    "Psnzed": "Aspfsd",
    "Psnzet": "Aspfsa",
    "Psnzem": "Aspfsl",
    "Psnzeo": "Aspfsi",
    "Psnzdi": "Aspfdn",
    "Psnzdr": "Aspfdg",
    "Psnzdd": "Aspfdd",
    "Psnzdt": "Aspfda",
    "Psnzdm": "Aspfdl",
    "Psnzdo": "Aspfdi",
    "Psnzmi": "Aspfpn",
    "Psnzmr": "Aspfpg",
    "Psnzmd": "Aspfpd",
    "Psnzmt": "Aspfpa",
    "Psnzmm": "Aspfpl",
    "Psnzmo": "Aspfpi",
    "Psnsei": "Aspnsn",
    "Psnser": "Aspnsg",
    "Psnsed": "Aspnsd",
    "Psnset": "Aspnsa",
    "Psnsem": "Aspnsl",
    "Psnseo": "Aspnsi",
    "Psnsdi": "Aspndn",
    "Psnsdr": "Aspndg",
    "Psnsdd": "Aspndd",
    "Psnsdt": "Aspnda",
    "Psnsdm": "Aspndl",
    "Psnsdo": "Aspndi",
    "Psnsmi": "Aspnpn",
    "Psnsmr": "Aspnpg",
    "Psnsmd": "Aspnpd",
    "Psnsmt": "Aspnpa",
    "Psnsmm": "Aspnpl",
    "Psnsmo": "Aspnpi",
    "Pdnmein": "Appmsnn",
    "Pdnmeid": "Appmsny",
    "Pdnmer": "Appmsg",
    "Pdnmed": "Appmsd",
    "Pdnmet": "Appmsa",
    "Pdnmetn": "Appmsan",
    "Pdnmetd": "Appmsay",
    "Pdnmem": "Appmsl",
    "Pdnmeo": "Appmsi",
    "Pdnmdi": "Appmdn",
    "Pdnmdr": "Appmdg",
    "Pdnmdd": "Appmdd",
    "Pdnmdt": "Appmda",
    "Pdnmdm": "Appmdl",
    "Pdnmdo": "Appmdi",
    "Pdnmmi": "Appmpn",
    "Pdnmmr": "Appmpg",
    "Pdnmmd": "Appmpd",
    "Pdnmmt": "Appmpa",
    "Pdnmmm": "Appmpl",
    "Pdnmmo": "Appmpi",
    "Pdnzei": "Appfsn",
    "Pdnzer": "Appfsg",
    "Pdnzed": "Appfsd",
    "Pdnzet": "Appfsa",
    "Pdnzem": "Appfsl",
    "Pdnzeo": "Appfsi",
    "Pdnzdi": "Appfdn",
    "Pdnzdr": "Appfdg",
    "Pdnzdd": "Appfdd",
    "Pdnzdt": "Appfda",
    "Pdnzdm": "Appfdl",
    "Pdnzdo": "Appfdi",
    "Pdnzmi": "Appfpn",
    "Pdnzmr": "Appfpg",
    "Pdnzmd": "Appfpd",
    "Pdnzmt": "Appfpa",
    "Pdnzmm": "Appfpl",
    "Pdnzmo": "Appfpi",
    "Pdnsei": "Appnsn",
    "Pdnser": "Appnsg",
    "Pdnsed": "Appnsd",
    "Pdnset": "Appnsa",
    "Pdnsem": "Appnsl",
    "Pdnseo": "Appnsi",
    "Pdnsdi": "Appndn",
    "Pdnsdr": "Appndg",
    "Pdnsdd": "Appndd",
    "Pdnsdt": "Appnda",
    "Pdnsdm": "Appndl",
    "Pdnsdo": "Appndi",
    "Pdnsmi": "Appnpn",
    "Pdnsmr": "Appnpg",
    "Pdnsmd": "Appnpd",
    "Pdnsmt": "Appnpa",
    "Pdnsmm": "Appnpl",
    "Pdnsmo": "Appnpi",
    "Rsn": "Rgp",
    "Rsr": "Rgc",
    "Rss": "Rgs",
    "Rd": "Rr",
    "Zop-ei": "Pp1-sn",
    "Zop-er": "Pp1-sg",
    "Zop-er--k": "Pp1-sg--y",
    "Zop-ed": "Pp1-sd",
    "Zop-ed--k": "Pp1-sd--y",
    "Zop-et": "Pp1-sa",
    "Zop-et--k": "Pp1-sa--y",
    "Zop-et--z": "Pp1-sa--b",
    "Zop-em": "Pp1-sl",
    "Zop-eo": "Pp1-si",
    "Zop-dr": "Pp1-dg",
    "Zop-dd": "Pp1-dd",
    "Zop-dt": "Pp1-da",
    "Zop-dm": "Pp1-dl",
    "Zop-do": "Pp1-di",
    "Zop-mr": "Pp1-pg",
    "Zop-md": "Pp1-pd",
    "Zop-mt": "Pp1-pa",
    "Zop-mm": "Pp1-pl",
    "Zop-mo": "Pp1-pi",
    "Zopmdi": "Pp1mdn",
    "Zopmmi": "Pp1mpn",
    "Zopzdi": "Pp1fdn",
    "Zopzmi": "Pp1fpn",
    "Zopsdi": "Pp1ndn",
    "Zopsmi": "Pp1npn",
    "Zod-ei": "Pp2-sn",
    "Zod-er": "Pp2-sg",
    "Zod-er--k": "Pp2-sg--y",
    "Zod-ed": "Pp2-sd",
    "Zod-ed--k": "Pp2-sd--y",
    "Zod-et": "Pp2-sa",
    "Zod-et--k": "Pp2-sa--y",
    "Zod-et--z": "Pp2-sa--b",
    "Zod-em": "Pp2-sl",
    "Zod-eo": "Pp2-si",
    "Zod-dr": "Pp2-dg",
    "Zod-dd": "Pp2-dd",
    "Zod-dt": "Pp2-da",
    "Zod-dm": "Pp2-dl",
    "Zod-do": "Pp2-di",
    "Zod-mr": "Pp2-pg",
    "Zod-md": "Pp2-pd",
    "Zod-mt": "Pp2-pa",
    "Zod-mm": "Pp2-pl",
    "Zod-mo": "Pp2-pi",
    "Zodmdi": "Pp2mdn",
    "Zodmmi": "Pp2mpn",
    "Zodzdi": "Pp2fdn",
    "Zodzmi": "Pp2fpn",
    "Zodsdi": "Pp2ndn",
    "Zodsmi": "Pp2npn",
    "Zotmei": "Pp3msn",
    "Zotmer": "Pp3msg",
    "Zotmer--k": "Pp3msg--y",
    "Zotmed": "Pp3msd",
    "Zotmed--k": "Pp3msd--y",
    "Zotmet": "Pp3msa",
    "Zotmet--k": "Pp3msa--y",
    "Zotmet--z": "Pp3msa--b",
    "Zotmem": "Pp3msl",
    "Zotmeo": "Pp3msi",
    "Zotmdi": "Pp3mdn",
    "Zotmdr": "Pp3mdg",
    "Zotmdr--k": "Pp3mdg--y",
    "Zotmdd": "Pp3mdd",
    "Zotmdd--k": "Pp3mdd--y",
    "Zotmdt": "Pp3mda",
    "Zotmdt--k": "Pp3mda--y",
    "Zotmdt--z": "Pp3mda--b",
    "Zotmdm": "Pp3mdl",
    "Zotmdo": "Pp3mdi",
    "Zotmmi": "Pp3mpn",
    "Zotmmr": "Pp3mpg",
    "Zotmmr--k": "Pp3mpg--y",
    "Zotmmd": "Pp3mpd",
    "Zotmmd--k": "Pp3mpd--y",
    "Zotmmt": "Pp3mpa",
    "Zotmmt--k": "Pp3mpa--y",
    "Zotmmt--z": "Pp3mpa--b",
    "Zotmmm": "Pp3mpl",
    "Zotmmo": "Pp3mpi",
    "Zotzei": "Pp3fsn",
    "Zotzer": "Pp3fsg",
    "Zotzer--k": "Pp3fsg--y",
    "Zotzed": "Pp3fsd",
    "Zotzed--k": "Pp3fsd--y",
    "Zotzet": "Pp3fsa",
    "Zotzet--k": "Pp3fsa--y",
    "Zotzet--z": "Pp3fsa--b",
    "Zotzem": "Pp3fsl",
    "Zotzeo": "Pp3fsi",
    "Zotzdi": "Pp3fdn",
    "Zotzdr": "Pp3fdg",
    "Zotzdr--k": "Pp3fdg--y",
    "Zotzdd": "Pp3fdd",
    "Zotzdd--k": "Pp3fdd--y",
    "Zotzdt": "Pp3fda",
    "Zotzdt--k": "Pp3fda--y",
    "Zotzdt--z": "Pp3fda--b",
    "Zotzdm": "Pp3fdl",
    "Zotzdo": "Pp3fdi",
    "Zotzmi": "Pp3fpn",
    "Zotzmr": "Pp3fpg",
    "Zotzmr--k": "Pp3fpg--y",
    "Zotzmd": "Pp3fpd",
    "Zotzmd--k": "Pp3fpd--y",
    "Zotzmt": "Pp3fpa",
    "Zotzmt--k": "Pp3fpa--y",
    "Zotzmt--z": "Pp3fpa--b",
    "Zotzmm": "Pp3fpl",
    "Zotzmo": "Pp3fpi",
    "Zotsei": "Pp3nsn",
    "Zotser": "Pp3nsg",
    "Zotser--k": "Pp3nsg--y",
    "Zotsed": "Pp3nsd",
    "Zotsed--k": "Pp3nsd--y",
    "Zotset": "Pp3nsa",
    "Zotset--k": "Pp3nsa--y",
    "Zotset--z": "Pp3nsa--b",
    "Zotsem": "Pp3nsl",
    "Zotseo": "Pp3nsi",
    "Zotsdi": "Pp3ndn",
    "Zotsdr": "Pp3ndg",
    "Zotsdr--k": "Pp3ndg--y",
    "Zotsdd": "Pp3ndd",
    "Zotsdd--k": "Pp3ndd--y",
    "Zotsdt": "Pp3nda",
    "Zotsdt--k": "Pp3nda--y",
    "Zotsdt--z": "Pp3nda--b",
    "Zotsdm": "Pp3ndl",
    "Zotsdo": "Pp3ndi",
    "Zotsmi": "Pp3npn",
    "Zotsmr": "Pp3npg",
    "Zotsmr--k": "Pp3npg--y",
    "Zotsmd": "Pp3npd",
    "Zotsmd--k": "Pp3npd--y",
    "Zotsmt": "Pp3npa",
    "Zotsmt--k": "Pp3npa--y",
    "Zotsmt--z": "Pp3npa--b",
    "Zotsmm": "Pp3npl",
    "Zotsmo": "Pp3npi",
    "Zspmeie": "Ps1msns",
    "Zspmeid": "Ps1msnd",
    "Zspmeim": "Ps1msnp",
    "Zspmere": "Ps1msgs",
    "Zspmerd": "Ps1msgd",
    "Zspmerm": "Ps1msgp",
    "Zspmede": "Ps1msds",
    "Zspmedd": "Ps1msdd",
    "Zspmedm": "Ps1msdp",
    "Zspmete": "Ps1msas",
    "Zspmetd": "Ps1msad",
    "Zspmetm": "Ps1msap",
    "Zspmeme": "Ps1msls",
    "Zspmemd": "Ps1msld",
    "Zspmemm": "Ps1mslp",
    "Zspmeoe": "Ps1msis",
    "Zspmeod": "Ps1msid",
    "Zspmeom": "Ps1msip",
    "Zspmdie": "Ps1mdns",
    "Zspmdid": "Ps1mdnd",
    "Zspmdim": "Ps1mdnp",
    "Zspmdre": "Ps1mdgs",
    "Zspmdrd": "Ps1mdgd",
    "Zspmdrm": "Ps1mdgp",
    "Zspmdde": "Ps1mdds",
    "Zspmddd": "Ps1mddd",
    "Zspmddm": "Ps1mddp",
    "Zspmdte": "Ps1mdas",
    "Zspmdtd": "Ps1mdad",
    "Zspmdtm": "Ps1mdap",
    "Zspmdme": "Ps1mdls",
    "Zspmdmd": "Ps1mdld",
    "Zspmdmm": "Ps1mdlp",
    "Zspmdoe": "Ps1mdis",
    "Zspmdod": "Ps1mdid",
    "Zspmdom": "Ps1mdip",
    "Zspmmie": "Ps1mpns",
    "Zspmmid": "Ps1mpnd",
    "Zspmmim": "Ps1mpnp",
    "Zspmmre": "Ps1mpgs",
    "Zspmmrd": "Ps1mpgd",
    "Zspmmrm": "Ps1mpgp",
    "Zspmmde": "Ps1mpds",
    "Zspmmdd": "Ps1mpdd",
    "Zspmmdm": "Ps1mpdp",
    "Zspmmte": "Ps1mpas",
    "Zspmmtd": "Ps1mpad",
    "Zspmmtm": "Ps1mpap",
    "Zspmmme": "Ps1mpls",
    "Zspmmmd": "Ps1mpld",
    "Zspmmmm": "Ps1mplp",
    "Zspmmoe": "Ps1mpis",
    "Zspmmod": "Ps1mpid",
    "Zspmmom": "Ps1mpip",
    "Zspzeie": "Ps1fsns",
    "Zspzeid": "Ps1fsnd",
    "Zspzeim": "Ps1fsnp",
    "Zspzere": "Ps1fsgs",
    "Zspzerd": "Ps1fsgd",
    "Zspzerm": "Ps1fsgp",
    "Zspzede": "Ps1fsds",
    "Zspzedd": "Ps1fsdd",
    "Zspzedm": "Ps1fsdp",
    "Zspzete": "Ps1fsas",
    "Zspzetd": "Ps1fsad",
    "Zspzetm": "Ps1fsap",
    "Zspzeme": "Ps1fsls",
    "Zspzemd": "Ps1fsld",
    "Zspzemm": "Ps1fslp",
    "Zspzeoe": "Ps1fsis",
    "Zspzeod": "Ps1fsid",
    "Zspzeom": "Ps1fsip",
    "Zspzdie": "Ps1fdns",
    "Zspzdid": "Ps1fdnd",
    "Zspzdim": "Ps1fdnp",
    "Zspzdre": "Ps1fdgs",
    "Zspzdrd": "Ps1fdgd",
    "Zspzdrm": "Ps1fdgp",
    "Zspzdde": "Ps1fdds",
    "Zspzddd": "Ps1fddd",
    "Zspzddm": "Ps1fddp",
    "Zspzdte": "Ps1fdas",
    "Zspzdtd": "Ps1fdad",
    "Zspzdtm": "Ps1fdap",
    "Zspzdme": "Ps1fdls",
    "Zspzdmd": "Ps1fdld",
    "Zspzdmm": "Ps1fdlp",
    "Zspzdoe": "Ps1fdis",
    "Zspzdod": "Ps1fdid",
    "Zspzdom": "Ps1fdip",
    "Zspzmie": "Ps1fpns",
    "Zspzmid": "Ps1fpnd",
    "Zspzmim": "Ps1fpnp",
    "Zspzmre": "Ps1fpgs",
    "Zspzmrd": "Ps1fpgd",
    "Zspzmrm": "Ps1fpgp",
    "Zspzmde": "Ps1fpds",
    "Zspzmdd": "Ps1fpdd",
    "Zspzmdm": "Ps1fpdp",
    "Zspzmte": "Ps1fpas",
    "Zspzmtd": "Ps1fpad",
    "Zspzmtm": "Ps1fpap",
    "Zspzmme": "Ps1fpls",
    "Zspzmmd": "Ps1fpld",
    "Zspzmmm": "Ps1fplp",
    "Zspzmoe": "Ps1fpis",
    "Zspzmod": "Ps1fpid",
    "Zspzmom": "Ps1fpip",
    "Zspseie": "Ps1nsns",
    "Zspseid": "Ps1nsnd",
    "Zspseim": "Ps1nsnp",
    "Zspsere": "Ps1nsgs",
    "Zspserd": "Ps1nsgd",
    "Zspserm": "Ps1nsgp",
    "Zspsede": "Ps1nsds",
    "Zspsedd": "Ps1nsdd",
    "Zspsedm": "Ps1nsdp",
    "Zspsete": "Ps1nsas",
    "Zspsetd": "Ps1nsad",
    "Zspsetm": "Ps1nsap",
    "Zspseme": "Ps1nsls",
    "Zspsemd": "Ps1nsld",
    "Zspsemm": "Ps1nslp",
    "Zspseoe": "Ps1nsis",
    "Zspseod": "Ps1nsid",
    "Zspseom": "Ps1nsip",
    "Zspsdie": "Ps1ndns",
    "Zspsdid": "Ps1ndnd",
    "Zspsdim": "Ps1ndnp",
    "Zspsdre": "Ps1ndgs",
    "Zspsdrd": "Ps1ndgd",
    "Zspsdrm": "Ps1ndgp",
    "Zspsdde": "Ps1ndds",
    "Zspsddd": "Ps1nddd",
    "Zspsddm": "Ps1nddp",
    "Zspsdte": "Ps1ndas",
    "Zspsdtd": "Ps1ndad",
    "Zspsdtm": "Ps1ndap",
    "Zspsdme": "Ps1ndls",
    "Zspsdmd": "Ps1ndld",
    "Zspsdmm": "Ps1ndlp",
    "Zspsdoe": "Ps1ndis",
    "Zspsdod": "Ps1ndid",
    "Zspsdom": "Ps1ndip",
    "Zspsmie": "Ps1npns",
    "Zspsmid": "Ps1npnd",
    "Zspsmim": "Ps1npnp",
    "Zspsmre": "Ps1npgs",
    "Zspsmrd": "Ps1npgd",
    "Zspsmrm": "Ps1npgp",
    "Zspsmde": "Ps1npds",
    "Zspsmdd": "Ps1npdd",
    "Zspsmdm": "Ps1npdp",
    "Zspsmte": "Ps1npas",
    "Zspsmtd": "Ps1npad",
    "Zspsmtm": "Ps1npap",
    "Zspsmme": "Ps1npls",
    "Zspsmmd": "Ps1npld",
    "Zspsmmm": "Ps1nplp",
    "Zspsmoe": "Ps1npis",
    "Zspsmod": "Ps1npid",
    "Zspsmom": "Ps1npip",
    "Zsdmeie": "Ps2msns",
    "Zsdmeid": "Ps2msnd",
    "Zsdmeim": "Ps2msnp",
    "Zsdmere": "Ps2msgs",
    "Zsdmerd": "Ps2msgd",
    "Zsdmerm": "Ps2msgp",
    "Zsdmede": "Ps2msds",
    "Zsdmedd": "Ps2msdd",
    "Zsdmedm": "Ps2msdp",
    "Zsdmete": "Ps2msas",
    "Zsdmetd": "Ps2msad",
    "Zsdmetm": "Ps2msap",
    "Zsdmeme": "Ps2msls",
    "Zsdmemd": "Ps2msld",
    "Zsdmemm": "Ps2mslp",
    "Zsdmeoe": "Ps2msis",
    "Zsdmeod": "Ps2msid",
    "Zsdmeom": "Ps2msip",
    "Zsdmdie": "Ps2mdns",
    "Zsdmdid": "Ps2mdnd",
    "Zsdmdim": "Ps2mdnp",
    "Zsdmdre": "Ps2mdgs",
    "Zsdmdrd": "Ps2mdgd",
    "Zsdmdrm": "Ps2mdgp",
    "Zsdmdde": "Ps2mdds",
    "Zsdmddd": "Ps2mddd",
    "Zsdmddm": "Ps2mddp",
    "Zsdmdte": "Ps2mdas",
    "Zsdmdtd": "Ps2mdad",
    "Zsdmdtm": "Ps2mdap",
    "Zsdmdme": "Ps2mdls",
    "Zsdmdmd": "Ps2mdld",
    "Zsdmdmm": "Ps2mdlp",
    "Zsdmdoe": "Ps2mdis",
    "Zsdmdod": "Ps2mdid",
    "Zsdmdom": "Ps2mdip",
    "Zsdmmie": "Ps2mpns",
    "Zsdmmid": "Ps2mpnd",
    "Zsdmmim": "Ps2mpnp",
    "Zsdmmre": "Ps2mpgs",
    "Zsdmmrd": "Ps2mpgd",
    "Zsdmmrm": "Ps2mpgp",
    "Zsdmmde": "Ps2mpds",
    "Zsdmmdd": "Ps2mpdd",
    "Zsdmmdm": "Ps2mpdp",
    "Zsdmmte": "Ps2mpas",
    "Zsdmmtd": "Ps2mpad",
    "Zsdmmtm": "Ps2mpap",
    "Zsdmmme": "Ps2mpls",
    "Zsdmmmd": "Ps2mpld",
    "Zsdmmmm": "Ps2mplp",
    "Zsdmmoe": "Ps2mpis",
    "Zsdmmod": "Ps2mpid",
    "Zsdmmom": "Ps2mpip",
    "Zsdzeie": "Ps2fsns",
    "Zsdzeid": "Ps2fsnd",
    "Zsdzeim": "Ps2fsnp",
    "Zsdzere": "Ps2fsgs",
    "Zsdzerd": "Ps2fsgd",
    "Zsdzerm": "Ps2fsgp",
    "Zsdzede": "Ps2fsds",
    "Zsdzedd": "Ps2fsdd",
    "Zsdzedm": "Ps2fsdp",
    "Zsdzete": "Ps2fsas",
    "Zsdzetd": "Ps2fsad",
    "Zsdzetm": "Ps2fsap",
    "Zsdzeme": "Ps2fsls",
    "Zsdzemd": "Ps2fsld",
    "Zsdzemm": "Ps2fslp",
    "Zsdzeoe": "Ps2fsis",
    "Zsdzeod": "Ps2fsid",
    "Zsdzeom": "Ps2fsip",
    "Zsdzdie": "Ps2fdns",
    "Zsdzdid": "Ps2fdnd",
    "Zsdzdim": "Ps2fdnp",
    "Zsdzdre": "Ps2fdgs",
    "Zsdzdrd": "Ps2fdgd",
    "Zsdzdrm": "Ps2fdgp",
    "Zsdzdde": "Ps2fdds",
    "Zsdzddd": "Ps2fddd",
    "Zsdzddm": "Ps2fddp",
    "Zsdzdte": "Ps2fdas",
    "Zsdzdtd": "Ps2fdad",
    "Zsdzdtm": "Ps2fdap",
    "Zsdzdme": "Ps2fdls",
    "Zsdzdmd": "Ps2fdld",
    "Zsdzdmm": "Ps2fdlp",
    "Zsdzdoe": "Ps2fdis",
    "Zsdzdod": "Ps2fdid",
    "Zsdzdom": "Ps2fdip",
    "Zsdzmie": "Ps2fpns",
    "Zsdzmid": "Ps2fpnd",
    "Zsdzmim": "Ps2fpnp",
    "Zsdzmre": "Ps2fpgs",
    "Zsdzmrd": "Ps2fpgd",
    "Zsdzmrm": "Ps2fpgp",
    "Zsdzmde": "Ps2fpds",
    "Zsdzmdd": "Ps2fpdd",
    "Zsdzmdm": "Ps2fpdp",
    "Zsdzmte": "Ps2fpas",
    "Zsdzmtd": "Ps2fpad",
    "Zsdzmtm": "Ps2fpap",
    "Zsdzmme": "Ps2fpls",
    "Zsdzmmd": "Ps2fpld",
    "Zsdzmmm": "Ps2fplp",
    "Zsdzmoe": "Ps2fpis",
    "Zsdzmod": "Ps2fpid",
    "Zsdzmom": "Ps2fpip",
    "Zsdseie": "Ps2nsns",
    "Zsdseid": "Ps2nsnd",
    "Zsdseim": "Ps2nsnp",
    "Zsdsere": "Ps2nsgs",
    "Zsdserd": "Ps2nsgd",
    "Zsdserm": "Ps2nsgp",
    "Zsdsede": "Ps2nsds",
    "Zsdsedd": "Ps2nsdd",
    "Zsdsedm": "Ps2nsdp",
    "Zsdsete": "Ps2nsas",
    "Zsdsetd": "Ps2nsad",
    "Zsdsetm": "Ps2nsap",
    "Zsdseme": "Ps2nsls",
    "Zsdsemd": "Ps2nsld",
    "Zsdsemm": "Ps2nslp",
    "Zsdseoe": "Ps2nsis",
    "Zsdseod": "Ps2nsid",
    "Zsdseom": "Ps2nsip",
    "Zsdsdie": "Ps2ndns",
    "Zsdsdid": "Ps2ndnd",
    "Zsdsdim": "Ps2ndnp",
    "Zsdsdre": "Ps2ndgs",
    "Zsdsdrd": "Ps2ndgd",
    "Zsdsdrm": "Ps2ndgp",
    "Zsdsdde": "Ps2ndds",
    "Zsdsddd": "Ps2nddd",
    "Zsdsddm": "Ps2nddp",
    "Zsdsdte": "Ps2ndas",
    "Zsdsdtd": "Ps2ndad",
    "Zsdsdtm": "Ps2ndap",
    "Zsdsdme": "Ps2ndls",
    "Zsdsdmd": "Ps2ndld",
    "Zsdsdmm": "Ps2ndlp",
    "Zsdsdoe": "Ps2ndis",
    "Zsdsdod": "Ps2ndid",
    "Zsdsdom": "Ps2ndip",
    "Zsdsmie": "Ps2npns",
    "Zsdsmid": "Ps2npnd",
    "Zsdsmim": "Ps2npnp",
    "Zsdsmre": "Ps2npgs",
    "Zsdsmrd": "Ps2npgd",
    "Zsdsmrm": "Ps2npgp",
    "Zsdsmde": "Ps2npds",
    "Zsdsmdd": "Ps2npdd",
    "Zsdsmdm": "Ps2npdp",
    "Zsdsmte": "Ps2npas",
    "Zsdsmtd": "Ps2npad",
    "Zsdsmtm": "Ps2npap",
    "Zsdsmme": "Ps2npls",
    "Zsdsmmd": "Ps2npld",
    "Zsdsmmm": "Ps2nplp",
    "Zsdsmoe": "Ps2npis",
    "Zsdsmod": "Ps2npid",
    "Zsdsmom": "Ps2npip",
    "Zstmeiem": "Ps3msnsm",
    "Zstmeiez": "Ps3msnsf",
    "Zstmeies": "Ps3msnsn",
    "Zstmeid": "Ps3msnd",
    "Zstmeim": "Ps3msnp",
    "Zstmerem": "Ps3msgsm",
    "Zstmerez": "Ps3msgsf",
    "Zstmeres": "Ps3msgsn",
    "Zstmerd": "Ps3msgd",
    "Zstmerm": "Ps3msgp",
    "Zstmedem": "Ps3msdsm",
    "Zstmedez": "Ps3msdsf",
    "Zstmedes": "Ps3msdsn",
    "Zstmedd": "Ps3msdd",
    "Zstmedm": "Ps3msdp",
    "Zstmetem": "Ps3msasm",
    "Zstmetez": "Ps3msasf",
    "Zstmetes": "Ps3msasn",
    "Zstmetd": "Ps3msad",
    "Zstmetm": "Ps3msap",
    "Zstmemem": "Ps3mslsm",
    "Zstmemez": "Ps3mslsf",
    "Zstmemes": "Ps3mslsn",
    "Zstmemd": "Ps3msld",
    "Zstmemm": "Ps3mslp",
    "Zstmeoem": "Ps3msism",
    "Zstmeoez": "Ps3msisf",
    "Zstmeoes": "Ps3msisn",
    "Zstmeod": "Ps3msid",
    "Zstmeom": "Ps3msip",
    "Zstmdiem": "Ps3mdnsm",
    "Zstmdiez": "Ps3mdnsf",
    "Zstmdies": "Ps3mdnsn",
    "Zstmdid": "Ps3mdnd",
    "Zstmdim": "Ps3mdnp",
    "Zstmdrem": "Ps3mdgsm",
    "Zstmdrez": "Ps3mdgsf",
    "Zstmdres": "Ps3mdgsn",
    "Zstmdrd": "Ps3mdgd",
    "Zstmdrm": "Ps3mdgp",
    "Zstmddem": "Ps3mddsm",
    "Zstmddez": "Ps3mddsf",
    "Zstmddes": "Ps3mddsn",
    "Zstmddd": "Ps3mddd",
    "Zstmddm": "Ps3mddp",
    "Zstmdtem": "Ps3mdasm",
    "Zstmdtez": "Ps3mdasf",
    "Zstmdtes": "Ps3mdasn",
    "Zstmdtd": "Ps3mdad",
    "Zstmdtm": "Ps3mdap",
    "Zstmdmem": "Ps3mdlsm",
    "Zstmdmez": "Ps3mdlsf",
    "Zstmdmes": "Ps3mdlsn",
    "Zstmdmd": "Ps3mdld",
    "Zstmdmm": "Ps3mdlp",
    "Zstmdoem": "Ps3mdism",
    "Zstmdoez": "Ps3mdisf",
    "Zstmdoes": "Ps3mdisn",
    "Zstmdod": "Ps3mdid",
    "Zstmdom": "Ps3mdip",
    "Zstmmiem": "Ps3mpnsm",
    "Zstmmiez": "Ps3mpnsf",
    "Zstmmies": "Ps3mpnsn",
    "Zstmmid": "Ps3mpnd",
    "Zstmmim": "Ps3mpnp",
    "Zstmmrem": "Ps3mpgsm",
    "Zstmmrez": "Ps3mpgsf",
    "Zstmmres": "Ps3mpgsn",
    "Zstmmrd": "Ps3mpgd",
    "Zstmmrm": "Ps3mpgp",
    "Zstmmdem": "Ps3mpdsm",
    "Zstmmdez": "Ps3mpdsf",
    "Zstmmdes": "Ps3mpdsn",
    "Zstmmdd": "Ps3mpdd",
    "Zstmmdm": "Ps3mpdp",
    "Zstmmtem": "Ps3mpasm",
    "Zstmmtez": "Ps3mpasf",
    "Zstmmtes": "Ps3mpasn",
    "Zstmmtd": "Ps3mpad",
    "Zstmmtm": "Ps3mpap",
    "Zstmmmem": "Ps3mplsm",
    "Zstmmmez": "Ps3mplsf",
    "Zstmmmes": "Ps3mplsn",
    "Zstmmmd": "Ps3mpld",
    "Zstmmmm": "Ps3mplp",
    "Zstmmoem": "Ps3mpism",
    "Zstmmoez": "Ps3mpisf",
    "Zstmmoes": "Ps3mpisn",
    "Zstmmod": "Ps3mpid",
    "Zstmmom": "Ps3mpip",
    "Zstzeiem": "Ps3fsnsm",
    "Zstzeiez": "Ps3fsnsf",
    "Zstzeies": "Ps3fsnsn",
    "Zstzeid": "Ps3fsnd",
    "Zstzeim": "Ps3fsnp",
    "Zstzerem": "Ps3fsgsm",
    "Zstzerez": "Ps3fsgsf",
    "Zstzeres": "Ps3fsgsn",
    "Zstzerd": "Ps3fsgd",
    "Zstzerm": "Ps3fsgp",
    "Zstzedem": "Ps3fsdsm",
    "Zstzedez": "Ps3fsdsf",
    "Zstzedes": "Ps3fsdsn",
    "Zstzedd": "Ps3fsdd",
    "Zstzedm": "Ps3fsdp",
    "Zstzetem": "Ps3fsasm",
    "Zstzetez": "Ps3fsasf",
    "Zstzetes": "Ps3fsasn",
    "Zstzetd": "Ps3fsad",
    "Zstzetm": "Ps3fsap",
    "Zstzemem": "Ps3fslsm",
    "Zstzemez": "Ps3fslsf",
    "Zstzemes": "Ps3fslsn",
    "Zstzemd": "Ps3fsld",
    "Zstzemm": "Ps3fslp",
    "Zstzeoem": "Ps3fsism",
    "Zstzeoez": "Ps3fsisf",
    "Zstzeoes": "Ps3fsisn",
    "Zstzeod": "Ps3fsid",
    "Zstzeom": "Ps3fsip",
    "Zstzdiem": "Ps3fdnsm",
    "Zstzdiez": "Ps3fdnsf",
    "Zstzdies": "Ps3fdnsn",
    "Zstzdid": "Ps3fdnd",
    "Zstzdim": "Ps3fdnp",
    "Zstzdrem": "Ps3fdgsm",
    "Zstzdrez": "Ps3fdgsf",
    "Zstzdres": "Ps3fdgsn",
    "Zstzdrd": "Ps3fdgd",
    "Zstzdrm": "Ps3fdgp",
    "Zstzddem": "Ps3fddsm",
    "Zstzddez": "Ps3fddsf",
    "Zstzddes": "Ps3fddsn",
    "Zstzddd": "Ps3fddd",
    "Zstzddm": "Ps3fddp",
    "Zstzdtem": "Ps3fdasm",
    "Zstzdtez": "Ps3fdasf",
    "Zstzdtes": "Ps3fdasn",
    "Zstzdtd": "Ps3fdad",
    "Zstzdtm": "Ps3fdap",
    "Zstzdmem": "Ps3fdlsm",
    "Zstzdmez": "Ps3fdlsf",
    "Zstzdmes": "Ps3fdlsn",
    "Zstzdmd": "Ps3fdld",
    "Zstzdmm": "Ps3fdlp",
    "Zstzdoem": "Ps3fdism",
    "Zstzdoez": "Ps3fdisf",
    "Zstzdoes": "Ps3fdisn",
    "Zstzdod": "Ps3fdid",
    "Zstzdom": "Ps3fdip",
    "Zstzmiem": "Ps3fpnsm",
    "Zstzmiez": "Ps3fpnsf",
    "Zstzmies": "Ps3fpnsn",
    "Zstzmid": "Ps3fpnd",
    "Zstzmim": "Ps3fpnp",
    "Zstzmrem": "Ps3fpgsm",
    "Zstzmrez": "Ps3fpgsf",
    "Zstzmres": "Ps3fpgsn",
    "Zstzmrd": "Ps3fpgd",
    "Zstzmrm": "Ps3fpgp",
    "Zstzmdem": "Ps3fpdsm",
    "Zstzmdez": "Ps3fpdsf",
    "Zstzmdes": "Ps3fpdsn",
    "Zstzmdd": "Ps3fpdd",
    "Zstzmdm": "Ps3fpdp",
    "Zstzmtem": "Ps3fpasm",
    "Zstzmtez": "Ps3fpasf",
    "Zstzmtes": "Ps3fpasn",
    "Zstzmtd": "Ps3fpad",
    "Zstzmtm": "Ps3fpap",
    "Zstzmmem": "Ps3fplsm",
    "Zstzmmez": "Ps3fplsf",
    "Zstzmmes": "Ps3fplsn",
    "Zstzmmd": "Ps3fpld",
    "Zstzmmm": "Ps3fplp",
    "Zstzmoem": "Ps3fpism",
    "Zstzmoez": "Ps3fpisf",
    "Zstzmoes": "Ps3fpisn",
    "Zstzmod": "Ps3fpid",
    "Zstzmom": "Ps3fpip",
    "Zstseiem": "Ps3nsnsm",
    "Zstseiez": "Ps3nsnsf",
    "Zstseies": "Ps3nsnsn",
    "Zstseid": "Ps3nsnd",
    "Zstseim": "Ps3nsnp",
    "Zstserem": "Ps3nsgsm",
    "Zstserez": "Ps3nsgsf",
    "Zstseres": "Ps3nsgsn",
    "Zstserd": "Ps3nsgd",
    "Zstserm": "Ps3nsgp",
    "Zstsedem": "Ps3nsdsm",
    "Zstsedez": "Ps3nsdsf",
    "Zstsedes": "Ps3nsdsn",
    "Zstsedd": "Ps3nsdd",
    "Zstsedm": "Ps3nsdp",
    "Zstsetem": "Ps3nsasm",
    "Zstsetez": "Ps3nsasf",
    "Zstsetes": "Ps3nsasn",
    "Zstsetd": "Ps3nsad",
    "Zstsetm": "Ps3nsap",
    "Zstsemem": "Ps3nslsm",
    "Zstsemez": "Ps3nslsf",
    "Zstsemes": "Ps3nslsn",
    "Zstsemd": "Ps3nsld",
    "Zstsemm": "Ps3nslp",
    "Zstseoem": "Ps3nsism",
    "Zstseoez": "Ps3nsisf",
    "Zstseoes": "Ps3nsisn",
    "Zstseod": "Ps3nsid",
    "Zstseom": "Ps3nsip",
    "Zstsdiem": "Ps3ndnsm",
    "Zstsdiez": "Ps3ndnsf",
    "Zstsdies": "Ps3ndnsn",
    "Zstsdid": "Ps3ndnd",
    "Zstsdim": "Ps3ndnp",
    "Zstsdrem": "Ps3ndgsm",
    "Zstsdrez": "Ps3ndgsf",
    "Zstsdres": "Ps3ndgsn",
    "Zstsdrd": "Ps3ndgd",
    "Zstsdrm": "Ps3ndgp",
    "Zstsddem": "Ps3nddsm",
    "Zstsddez": "Ps3nddsf",
    "Zstsddes": "Ps3nddsn",
    "Zstsddd": "Ps3nddd",
    "Zstsddm": "Ps3nddp",
    "Zstsdtem": "Ps3ndasm",
    "Zstsdtez": "Ps3ndasf",
    "Zstsdtes": "Ps3ndasn",
    "Zstsdtd": "Ps3ndad",
    "Zstsdtm": "Ps3ndap",
    "Zstsdmem": "Ps3ndlsm",
    "Zstsdmez": "Ps3ndlsf",
    "Zstsdmes": "Ps3ndlsn",
    "Zstsdmd": "Ps3ndld",
    "Zstsdmm": "Ps3ndlp",
    "Zstsdoem": "Ps3ndism",
    "Zstsdoez": "Ps3ndisf",
    "Zstsdoes": "Ps3ndisn",
    "Zstsdod": "Ps3ndid",
    "Zstsdom": "Ps3ndip",
    "Zstsmiem": "Ps3npnsm",
    "Zstsmiez": "Ps3npnsf",
    "Zstsmies": "Ps3npnsn",
    "Zstsmid": "Ps3npnd",
    "Zstsmim": "Ps3npnp",
    "Zstsmrem": "Ps3npgsm",
    "Zstsmrez": "Ps3npgsf",
    "Zstsmres": "Ps3npgsn",
    "Zstsmrd": "Ps3npgd",
    "Zstsmrm": "Ps3npgp",
    "Zstsmdem": "Ps3npdsm",
    "Zstsmdez": "Ps3npdsf",
    "Zstsmdes": "Ps3npdsn",
    "Zstsmdd": "Ps3npdd",
    "Zstsmdm": "Ps3npdp",
    "Zstsmtem": "Ps3npasm",
    "Zstsmtez": "Ps3npasf",
    "Zstsmtes": "Ps3npasn",
    "Zstsmtd": "Ps3npad",
    "Zstsmtm": "Ps3npap",
    "Zstsmmem": "Ps3nplsm",
    "Zstsmmez": "Ps3nplsf",
    "Zstsmmes": "Ps3nplsn",
    "Zstsmmd": "Ps3npld",
    "Zstsmmm": "Ps3nplp",
    "Zstsmoem": "Ps3npism",
    "Zstsmoez": "Ps3npisf",
    "Zstsmoes": "Ps3npisn",
    "Zstsmod": "Ps3npid",
    "Zstsmom": "Ps3npip",
    "Zk-mei": "Pd-msn",
    "Zk-mer": "Pd-msg",
    "Zk-med": "Pd-msd",
    "Zk-met": "Pd-msa",
    "Zk-mem": "Pd-msl",
    "Zk-meo": "Pd-msi",
    "Zk-mdi": "Pd-mdn",
    "Zk-mdr": "Pd-mdg",
    "Zk-mdd": "Pd-mdd",
    "Zk-mdt": "Pd-mda",
    "Zk-mdm": "Pd-mdl",
    "Zk-mdo": "Pd-mdi",
    "Zk-mmi": "Pd-mpn",
    "Zk-mmr": "Pd-mpg",
    "Zk-mmd": "Pd-mpd",
    "Zk-mmt": "Pd-mpa",
    "Zk-mmm": "Pd-mpl",
    "Zk-mmo": "Pd-mpi",
    "Zk-zei": "Pd-fsn",
    "Zk-zer": "Pd-fsg",
    "Zk-zed": "Pd-fsd",
    "Zk-zet": "Pd-fsa",
    "Zk-zem": "Pd-fsl",
    "Zk-zeo": "Pd-fsi",
    "Zk-zdi": "Pd-fdn",
    "Zk-zdr": "Pd-fdg",
    "Zk-zdd": "Pd-fdd",
    "Zk-zdt": "Pd-fda",
    "Zk-zdm": "Pd-fdl",
    "Zk-zdo": "Pd-fdi",
    "Zk-zmi": "Pd-fpn",
    "Zk-zmr": "Pd-fpg",
    "Zk-zmd": "Pd-fpd",
    "Zk-zmt": "Pd-fpa",
    "Zk-zmm": "Pd-fpl",
    "Zk-zmo": "Pd-fpi",
    "Zk-sei": "Pd-nsn",
    "Zk-ser": "Pd-nsg",
    "Zk-sed": "Pd-nsd",
    "Zk-set": "Pd-nsa",
    "Zk-sem": "Pd-nsl",
    "Zk-seo": "Pd-nsi",
    "Zk-sdi": "Pd-ndn",
    "Zk-sdr": "Pd-ndg",
    "Zk-sdd": "Pd-ndd",
    "Zk-sdt": "Pd-nda",
    "Zk-sdm": "Pd-ndl",
    "Zk-sdo": "Pd-ndi",
    "Zk-smi": "Pd-npn",
    "Zk-smr": "Pd-npg",
    "Zk-smd": "Pd-npd",
    "Zk-smt": "Pd-npa",
    "Zk-smm": "Pd-npl",
    "Zk-smo": "Pd-npi",
    "Zz": "Pr",
    "Zz----em": "Pr----sm",
    "Zz-mei": "Pr-msn",
    "Zz-mer": "Pr-msg",
    "Zz-med": "Pr-msd",
    "Zz-met": "Pr-msa",
    "Zz-mem": "Pr-msl",
    "Zz-meo": "Pr-msi",
    "Zz-mdi": "Pr-mdn",
    "Zz-mdr": "Pr-mdg",
    "Zz-mdd": "Pr-mdd",
    "Zz-mdt": "Pr-mda",
    "Zz-mdm": "Pr-mdl",
    "Zz-mdo": "Pr-mdi",
    "Zz-mmi": "Pr-mpn",
    "Zz-mmr": "Pr-mpg",
    "Zz-mmd": "Pr-mpd",
    "Zz-mmt": "Pr-mpa",
    "Zz-mmm": "Pr-mpl",
    "Zz-mmo": "Pr-mpi",
    "Zz-zei": "Pr-fsn",
    "Zz-zer": "Pr-fsg",
    "Zz-zed": "Pr-fsd",
    "Zz-zet": "Pr-fsa",
    "Zz-zem": "Pr-fsl",
    "Zz-zeo": "Pr-fsi",
    "Zz-zdi": "Pr-fdn",
    "Zz-zdr": "Pr-fdg",
    "Zz-zdd": "Pr-fdd",
    "Zz-zdt": "Pr-fda",
    "Zz-zdm": "Pr-fdl",
    "Zz-zdo": "Pr-fdi",
    "Zz-zmi": "Pr-fpn",
    "Zz-zmr": "Pr-fpg",
    "Zz-zmd": "Pr-fpd",
    "Zz-zmt": "Pr-fpa",
    "Zz-zmm": "Pr-fpl",
    "Zz-zmo": "Pr-fpi",
    "Zz-sei": "Pr-nsn",
    "Zz-ser": "Pr-nsg",
    "Zz-sed": "Pr-nsd",
    "Zz-set": "Pr-nsa",
    "Zz-sem": "Pr-nsl",
    "Zz-seo": "Pr-nsi",
    "Zz-sdi": "Pr-ndn",
    "Zz-sdr": "Pr-ndg",
    "Zz-sdd": "Pr-ndd",
    "Zz-sdt": "Pr-nda",
    "Zz-sdm": "Pr-ndl",
    "Zz-sdo": "Pr-ndi",
    "Zz-smi": "Pr-npn",
    "Zz-smr": "Pr-npg",
    "Zz-smd": "Pr-npd",
    "Zz-smt": "Pr-npa",
    "Zz-smm": "Pr-npl",
    "Zz-smo": "Pr-npi",
    "Zp------k": "Px------y",
    "Zp---r": "Px---g",
    "Zp---d": "Px---d",
    "Zp---d--k": "Px---d--y",
    "Zp---t": "Px---a",
    "Zp---t--z": "Px---a--b",
    "Zp---m": "Px---l",
    "Zp---o": "Px---i",
    "Zp-mei": "Px-msn",
    "Zp-mer": "Px-msg",
    "Zp-med": "Px-msd",
    "Zp-met": "Px-msa",
    "Zp-mem": "Px-msl",
    "Zp-meo": "Px-msi",
    "Zp-mdi": "Px-mdn",
    "Zp-mdr": "Px-mdg",
    "Zp-mdd": "Px-mdd",
    "Zp-mdt": "Px-mda",
    "Zp-mdm": "Px-mdl",
    "Zp-mdo": "Px-mdi",
    "Zp-mmi": "Px-mpn",
    "Zp-mmr": "Px-mpg",
    "Zp-mmd": "Px-mpd",
    "Zp-mmt": "Px-mpa",
    "Zp-mmm": "Px-mpl",
    "Zp-mmo": "Px-mpi",
    "Zp-zei": "Px-fsn",
    "Zp-zer": "Px-fsg",
    "Zp-zed": "Px-fsd",
    "Zp-zet": "Px-fsa",
    "Zp-zem": "Px-fsl",
    "Zp-zeo": "Px-fsi",
    "Zp-zdi": "Px-fdn",
    "Zp-zdr": "Px-fdg",
    "Zp-zdd": "Px-fdd",
    "Zp-zdt": "Px-fda",
    "Zp-zdm": "Px-fdl",
    "Zp-zdo": "Px-fdi",
    "Zp-zmi": "Px-fpn",
    "Zp-zmr": "Px-fpg",
    "Zp-zmd": "Px-fpd",
    "Zp-zmt": "Px-fpa",
    "Zp-zmm": "Px-fpl",
    "Zp-zmo": "Px-fpi",
    "Zp-sei": "Px-nsn",
    "Zp-ser": "Px-nsg",
    "Zp-sed": "Px-nsd",
    "Zp-set": "Px-nsa",
    "Zp-sem": "Px-nsl",
    "Zp-seo": "Px-nsi",
    "Zp-sdi": "Px-ndn",
    "Zp-sdr": "Px-ndg",
    "Zp-sdd": "Px-ndd",
    "Zp-sdt": "Px-nda",
    "Zp-sdm": "Px-ndl",
    "Zp-sdo": "Px-ndi",
    "Zp-smi": "Px-npn",
    "Zp-smr": "Px-npg",
    "Zp-smd": "Px-npd",
    "Zp-smt": "Px-npa",
    "Zp-smm": "Px-npl",
    "Zp-smo": "Px-npi",
    "Zc-mei": "Pg-msn",
    "Zc-mer": "Pg-msg",
    "Zc-med": "Pg-msd",
    "Zc-met": "Pg-msa",
    "Zc-mem": "Pg-msl",
    "Zc-meo": "Pg-msi",
    "Zc-mdi": "Pg-mdn",
    "Zc-mdr": "Pg-mdg",
    "Zc-mdd": "Pg-mdd",
    "Zc-mdt": "Pg-mda",
    "Zc-mdm": "Pg-mdl",
    "Zc-mdo": "Pg-mdi",
    "Zc-mmi": "Pg-mpn",
    "Zc-mmr": "Pg-mpg",
    "Zc-mmd": "Pg-mpd",
    "Zc-mmt": "Pg-mpa",
    "Zc-mmm": "Pg-mpl",
    "Zc-mmo": "Pg-mpi",
    "Zc-zei": "Pg-fsn",
    "Zc-zer": "Pg-fsg",
    "Zc-zed": "Pg-fsd",
    "Zc-zet": "Pg-fsa",
    "Zc-zem": "Pg-fsl",
    "Zc-zeo": "Pg-fsi",
    "Zc-zdi": "Pg-fdn",
    "Zc-zdr": "Pg-fdg",
    "Zc-zdd": "Pg-fdd",
    "Zc-zdt": "Pg-fda",
    "Zc-zdm": "Pg-fdl",
    "Zc-zdo": "Pg-fdi",
    "Zc-zmi": "Pg-fpn",
    "Zc-zmr": "Pg-fpg",
    "Zc-zmd": "Pg-fpd",
    "Zc-zmt": "Pg-fpa",
    "Zc-zmm": "Pg-fpl",
    "Zc-zmo": "Pg-fpi",
    "Zc-sei": "Pg-nsn",
    "Zc-ser": "Pg-nsg",
    "Zc-sed": "Pg-nsd",
    "Zc-set": "Pg-nsa",
    "Zc-sem": "Pg-nsl",
    "Zc-seo": "Pg-nsi",
    "Zc-sdi": "Pg-ndn",
    "Zc-sdr": "Pg-ndg",
    "Zc-sdd": "Pg-ndd",
    "Zc-sdt": "Pg-nda",
    "Zc-sdm": "Pg-ndl",
    "Zc-sdo": "Pg-ndi",
    "Zc-smi": "Pg-npn",
    "Zc-smr": "Pg-npg",
    "Zc-smd": "Pg-npd",
    "Zc-smt": "Pg-npa",
    "Zc-smm": "Pg-npl",
    "Zc-smo": "Pg-npi",
    "Zv----em": "Pq----sm",
    "Zv----ez": "Pq----sf",
    "Zv----es": "Pq----sn",
    "Zv----d": "Pq----d",
    "Zv----m": "Pq----p",
    "Zv-mei": "Pq-msn",
    "Zv-mer": "Pq-msg",
    "Zv-med": "Pq-msd",
    "Zv-met": "Pq-msa",
    "Zv-mem": "Pq-msl",
    "Zv-meo": "Pq-msi",
    "Zv-mdi": "Pq-mdn",
    "Zv-mdr": "Pq-mdg",
    "Zv-mdd": "Pq-mdd",
    "Zv-mdt": "Pq-mda",
    "Zv-mdm": "Pq-mdl",
    "Zv-mdo": "Pq-mdi",
    "Zv-mmi": "Pq-mpn",
    "Zv-mmr": "Pq-mpg",
    "Zv-mmd": "Pq-mpd",
    "Zv-mmt": "Pq-mpa",
    "Zv-mmm": "Pq-mpl",
    "Zv-mmo": "Pq-mpi",
    "Zv-zei": "Pq-fsn",
    "Zv-zer": "Pq-fsg",
    "Zv-zed": "Pq-fsd",
    "Zv-zet": "Pq-fsa",
    "Zv-zem": "Pq-fsl",
    "Zv-zeo": "Pq-fsi",
    "Zv-zdi": "Pq-fdn",
    "Zv-zdr": "Pq-fdg",
    "Zv-zdd": "Pq-fdd",
    "Zv-zdt": "Pq-fda",
    "Zv-zdm": "Pq-fdl",
    "Zv-zdo": "Pq-fdi",
    "Zv-zmi": "Pq-fpn",
    "Zv-zmr": "Pq-fpg",
    "Zv-zmd": "Pq-fpd",
    "Zv-zmt": "Pq-fpa",
    "Zv-zmm": "Pq-fpl",
    "Zv-zmo": "Pq-fpi",
    "Zv-sei": "Pq-nsn",
    "Zv-ser": "Pq-nsg",
    "Zv-sed": "Pq-nsd",
    "Zv-set": "Pq-nsa",
    "Zv-sem": "Pq-nsl",
    "Zv-seo": "Pq-nsi",
    "Zv-sdi": "Pq-ndn",
    "Zv-sdr": "Pq-ndg",
    "Zv-sdd": "Pq-ndd",
    "Zv-sdt": "Pq-nda",
    "Zv-sdm": "Pq-ndl",
    "Zv-sdo": "Pq-ndi",
    "Zv-smi": "Pq-npn",
    "Zv-smr": "Pq-npg",
    "Zv-smd": "Pq-npd",
    "Zv-smt": "Pq-npa",
    "Zv-smm": "Pq-npl",
    "Zv-smo": "Pq-npi",
    "Zn-mei": "Pi-msn",
    "Zn-mer": "Pi-msg",
    "Zn-med": "Pi-msd",
    "Zn-met": "Pi-msa",
    "Zn-mem": "Pi-msl",
    "Zn-meo": "Pi-msi",
    "Zn-mdi": "Pi-mdn",
    "Zn-mdr": "Pi-mdg",
    "Zn-mdd": "Pi-mdd",
    "Zn-mdt": "Pi-mda",
    "Zn-mdm": "Pi-mdl",
    "Zn-mdo": "Pi-mdi",
    "Zn-mmi": "Pi-mpn",
    "Zn-mmr": "Pi-mpg",
    "Zn-mmd": "Pi-mpd",
    "Zn-mmt": "Pi-mpa",
    "Zn-mmm": "Pi-mpl",
    "Zn-mmo": "Pi-mpi",
    "Zn-zei": "Pi-fsn",
    "Zn-zer": "Pi-fsg",
    "Zn-zed": "Pi-fsd",
    "Zn-zet": "Pi-fsa",
    "Zn-zem": "Pi-fsl",
    "Zn-zeo": "Pi-fsi",
    "Zn-zdi": "Pi-fdn",
    "Zn-zdr": "Pi-fdg",
    "Zn-zdd": "Pi-fdd",
    "Zn-zdt": "Pi-fda",
    "Zn-zdm": "Pi-fdl",
    "Zn-zdo": "Pi-fdi",
    "Zn-zmi": "Pi-fpn",
    "Zn-zmr": "Pi-fpg",
    "Zn-zmd": "Pi-fpd",
    "Zn-zmt": "Pi-fpa",
    "Zn-zmm": "Pi-fpl",
    "Zn-zmo": "Pi-fpi",
    "Zn-sei": "Pi-nsn",
    "Zn-ser": "Pi-nsg",
    "Zn-sed": "Pi-nsd",
    "Zn-set": "Pi-nsa",
    "Zn-sem": "Pi-nsl",
    "Zn-seo": "Pi-nsi",
    "Zn-sdi": "Pi-ndn",
    "Zn-sdr": "Pi-ndg",
    "Zn-sdd": "Pi-ndd",
    "Zn-sdt": "Pi-nda",
    "Zn-sdm": "Pi-ndl",
    "Zn-sdo": "Pi-ndi",
    "Zn-smi": "Pi-npn",
    "Zn-smr": "Pi-npg",
    "Zn-smd": "Pi-npd",
    "Zn-smt": "Pi-npa",
    "Zn-smm": "Pi-npl",
    "Zn-smo": "Pi-npi",
    "Zl-mei": "Pz-msn",
    "Zl-mer": "Pz-msg",
    "Zl-med": "Pz-msd",
    "Zl-met": "Pz-msa",
    "Zl-mem": "Pz-msl",
    "Zl-meo": "Pz-msi",
    "Zl-mdi": "Pz-mdn",
    "Zl-mdr": "Pz-mdg",
    "Zl-mdd": "Pz-mdd",
    "Zl-mdt": "Pz-mda",
    "Zl-mdm": "Pz-mdl",
    "Zl-mdo": "Pz-mdi",
    "Zl-mmi": "Pz-mpn",
    "Zl-mmr": "Pz-mpg",
    "Zl-mmd": "Pz-mpd",
    "Zl-mmt": "Pz-mpa",
    "Zl-mmm": "Pz-mpl",
    "Zl-mmo": "Pz-mpi",
    "Zl-zei": "Pz-fsn",
    "Zl-zer": "Pz-fsg",
    "Zl-zed": "Pz-fsd",
    "Zl-zet": "Pz-fsa",
    "Zl-zem": "Pz-fsl",
    "Zl-zeo": "Pz-fsi",
    "Zl-zdi": "Pz-fdn",
    "Zl-zdr": "Pz-fdg",
    "Zl-zdd": "Pz-fdd",
    "Zl-zdt": "Pz-fda",
    "Zl-zdm": "Pz-fdl",
    "Zl-zdo": "Pz-fdi",
    "Zl-zmi": "Pz-fpn",
    "Zl-zmr": "Pz-fpg",
    "Zl-zmd": "Pz-fpd",
    "Zl-zmt": "Pz-fpa",
    "Zl-zmm": "Pz-fpl",
    "Zl-zmo": "Pz-fpi",
    "Zl-sei": "Pz-nsn",
    "Zl-ser": "Pz-nsg",
    "Zl-sed": "Pz-nsd",
    "Zl-set": "Pz-nsa",
    "Zl-sem": "Pz-nsl",
    "Zl-seo": "Pz-nsi",
    "Zl-sdi": "Pz-ndn",
    "Zl-sdr": "Pz-ndg",
    "Zl-sdd": "Pz-ndd",
    "Zl-sdt": "Pz-nda",
    "Zl-sdm": "Pz-ndl",
    "Zl-sdo": "Pz-ndi",
    "Zl-smi": "Pz-npn",
    "Zl-smr": "Pz-npg",
    "Zl-smd": "Pz-npd",
    "Zl-smt": "Pz-npa",
    "Zl-smm": "Pz-npl",
    "Zl-smo": "Pz-npi",
    "Kag": "Mdc",
    "Kav": "Mdo",
    "Krg": "Mrc",
    "Krv": "Mro",
    "Kbg-mi": "Mlc-pn",
    "Kbg-mr": "Mlc-pg",
    "Kbg-md": "Mlc-pd",
    "Kbg-mt": "Mlc-pa",
    "Kbg-mm": "Mlc-pl",
    "Kbg-mo": "Mlc-pi",
    "Kbgmdi": "Mlcmdn",
    "Kbgmdr": "Mlcmdg",
    "Kbgmdd": "Mlcmdd",
    "Kbgmdt": "Mlcmda",
    "Kbgmdm": "Mlcmdl",
    "Kbgmdo": "Mlcmdi",
    "Kbgmmi": "Mlcmpn",
    "Kbgmmr": "Mlcmpg",
    "Kbgmmd": "Mlcmpd",
    "Kbgmmt": "Mlcmpa",
    "Kbgmmm": "Mlcmpl",
    "Kbgmmo": "Mlcmpi",
    "Kbgzdi": "Mlcfdn",
    "Kbgzdr": "Mlcfdg",
    "Kbgzdd": "Mlcfdd",
    "Kbgzdt": "Mlcfda",
    "Kbgzdm": "Mlcfdl",
    "Kbgzdo": "Mlcfdi",
    "Kbgzmi": "Mlcfpn",
    "Kbgzmr": "Mlcfpg",
    "Kbgzmd": "Mlcfpd",
    "Kbgzmt": "Mlcfpa",
    "Kbgzmm": "Mlcfpl",
    "Kbgzmo": "Mlcfpi",
    "Kbgsdi": "Mlcndn",
    "Kbgsdr": "Mlcndg",
    "Kbgsdd": "Mlcndd",
    "Kbgsdt": "Mlcnda",
    "Kbgsdm": "Mlcndl",
    "Kbgsdo": "Mlcndi",
    "Kbgsmi": "Mlcnpn",
    "Kbgsmr": "Mlcnpg",
    "Kbgsmd": "Mlcnpd",
    "Kbgsmt": "Mlcnpa",
    "Kbgsmm": "Mlcnpl",
    "Kbgsmo": "Mlcnpi",
    "Kbvmei": "Mlomsn",
    "Kbvmer": "Mlomsg",
    "Kbvmed": "Mlomsd",
    "Kbvmet": "Mlomsa",
    "Kbvmem": "Mlomsl",
    "Kbvmeo": "Mlomsi",
    "Kbvmdi": "Mlomdn",
    "Kbvmdr": "Mlomdg",
    "Kbvmdd": "Mlomdd",
    "Kbvmdt": "Mlomda",
    "Kbvmdm": "Mlomdl",
    "Kbvmdo": "Mlomdi",
    "Kbvmmi": "Mlompn",
    "Kbvmmr": "Mlompg",
    "Kbvmmd": "Mlompd",
    "Kbvmmt": "Mlompa",
    "Kbvmmm": "Mlompl",
    "Kbvmmo": "Mlompi",
    "Kbvzei": "Mlofsn",
    "Kbvzer": "Mlofsg",
    "Kbvzed": "Mlofsd",
    "Kbvzet": "Mlofsa",
    "Kbvzem": "Mlofsl",
    "Kbvzeo": "Mlofsi",
    "Kbvzdi": "Mlofdn",
    "Kbvzdr": "Mlofdg",
    "Kbvzdd": "Mlofdd",
    "Kbvzdt": "Mlofda",
    "Kbvzdm": "Mlofdl",
    "Kbvzdo": "Mlofdi",
    "Kbvzmi": "Mlofpn",
    "Kbvzmr": "Mlofpg",
    "Kbvzmd": "Mlofpd",
    "Kbvzmt": "Mlofpa",
    "Kbvzmm": "Mlofpl",
    "Kbvzmo": "Mlofpi",
    "Kbvsei": "Mlonsn",
    "Kbvser": "Mlonsg",
    "Kbvsed": "Mlonsd",
    "Kbvset": "Mlonsa",
    "Kbvsem": "Mlonsl",
    "Kbvseo": "Mlonsi",
    "Kbvsdi": "Mlondn",
    "Kbvsdr": "Mlondg",
    "Kbvsdd": "Mlondd",
    "Kbvsdt": "Mlonda",
    "Kbvsdm": "Mlondl",
    "Kbvsdo": "Mlondi",
    "Kbvsmi": "Mlonpn",
    "Kbvsmr": "Mlonpg",
    "Kbvsmd": "Mlonpd",
    "Kbvsmt": "Mlonpa",
    "Kbvsmm": "Mlonpl",
    "Kbvsmo": "Mlonpi",
    "Kbzmei": "Mlpmsn",
    "Kbzmein": "Mlpmsnn",
    "Kbzmeid": "Mlpmsny",
    "Kbzmer": "Mlpmsg",
    "Kbzmed": "Mlpmsd",
    "Kbzmet": "Mlpmsa",
    "Kbzmetn": "Mlpmsan",
    "Kbzmetd": "Mlpmsay",
    "Kbzmem": "Mlpmsl",
    "Kbzmeo": "Mlpmsi",
    "Kbzmdi": "Mlpmdn",
    "Kbzmdr": "Mlpmdg",
    "Kbzmdd": "Mlpmdd",
    "Kbzmdt": "Mlpmda",
    "Kbzmdm": "Mlpmdl",
    "Kbzmdo": "Mlpmdi",
    "Kbzmmi": "Mlpmpn",
    "Kbzmmr": "Mlpmpg",
    "Kbzmmd": "Mlpmpd",
    "Kbzmmt": "Mlpmpa",
    "Kbzmmm": "Mlpmpl",
    "Kbzmmo": "Mlpmpi",
    "Kbzzei": "Mlpfsn",
    "Kbzzer": "Mlpfsg",
    "Kbzzed": "Mlpfsd",
    "Kbzzet": "Mlpfsa",
    "Kbzzem": "Mlpfsl",
    "Kbzzeo": "Mlpfsi",
    "Kbzzdi": "Mlpfdn",
    "Kbzzdr": "Mlpfdg",
    "Kbzzdd": "Mlpfdd",
    "Kbzzdt": "Mlpfda",
    "Kbzzdm": "Mlpfdl",
    "Kbzzdo": "Mlpfdi",
    "Kbzzmi": "Mlpfpn",
    "Kbzzmr": "Mlpfpg",
    "Kbzzmd": "Mlpfpd",
    "Kbzzmt": "Mlpfpa",
    "Kbzzmm": "Mlpfpl",
    "Kbzzmo": "Mlpfpi",
    "Kbzsei": "Mlpnsn",
    "Kbzser": "Mlpnsg",
    "Kbzsed": "Mlpnsd",
    "Kbzset": "Mlpnsa",
    "Kbzsem": "Mlpnsl",
    "Kbzseo": "Mlpnsi",
    "Kbzsdi": "Mlpndn",
    "Kbzsdr": "Mlpndg",
    "Kbzsdd": "Mlpndd",
    "Kbzsdt": "Mlpnda",
    "Kbzsdm": "Mlpndl",
    "Kbzsdo": "Mlpndi",
    "Kbzsmi": "Mlpnpn",
    "Kbzsmr": "Mlpnpg",
    "Kbzsmd": "Mlpnpd",
    "Kbzsmt": "Mlpnpa",
    "Kbzsmm": "Mlpnpl",
    "Kbzsmo": "Mlpnpi",
    "Kbdmein": "Mlsmsnn",
    "Kbdmeid": "Mlsmsny",
    "Kbdmer": "Mlsmsg",
    "Kbdmed": "Mlsmsd",
    "Kbdmet": "Mlsmsa",
    "Kbdmetn": "Mlsmsan",
    "Kbdmetd": "Mlsmsay",
    "Kbdmem": "Mlsmsl",
    "Kbdmeo": "Mlsmsi",
    "Kbdmdi": "Mlsmdn",
    "Kbdmdr": "Mlsmdg",
    "Kbdmdd": "Mlsmdd",
    "Kbdmdt": "Mlsmda",
    "Kbdmdm": "Mlsmdl",
    "Kbdmdo": "Mlsmdi",
    "Kbdmmi": "Mlsmpn",
    "Kbdmmr": "Mlsmpg",
    "Kbdmmd": "Mlsmpd",
    "Kbdmmt": "Mlsmpa",
    "Kbdmmm": "Mlsmpl",
    "Kbdmmo": "Mlsmpi",
    "Kbdzei": "Mlsfsn",
    "Kbdzer": "Mlsfsg",
    "Kbdzed": "Mlsfsd",
    "Kbdzet": "Mlsfsa",
    "Kbdzem": "Mlsfsl",
    "Kbdzeo": "Mlsfsi",
    "Kbdzdi": "Mlsfdn",
    "Kbdzdr": "Mlsfdg",
    "Kbdzdd": "Mlsfdd",
    "Kbdzdt": "Mlsfda",
    "Kbdzdm": "Mlsfdl",
    "Kbdzdo": "Mlsfdi",
    "Kbdzmi": "Mlsfpn",
    "Kbdzmr": "Mlsfpg",
    "Kbdzmd": "Mlsfpd",
    "Kbdzmt": "Mlsfpa",
    "Kbdzmm": "Mlsfpl",
    "Kbdzmo": "Mlsfpi",
    "Kbdsei": "Mlsnsn",
    "Kbdser": "Mlsnsg",
    "Kbdsed": "Mlsnsd",
    "Kbdset": "Mlsnsa",
    "Kbdsem": "Mlsnsl",
    "Kbdseo": "Mlsnsi",
    "Kbdsdi": "Mlsndn",
    "Kbdsdr": "Mlsndg",
    "Kbdsdd": "Mlsndd",
    "Kbdsdt": "Mlsnda",
    "Kbdsdm": "Mlsndl",
    "Kbdsdo": "Mlsndi",
    "Kbdsmi": "Mlsnpn",
    "Kbdsmr": "Mlsnpg",
    "Kbdsmd": "Mlsnpd",
    "Kbdsmt": "Mlsnpa",
    "Kbdsmm": "Mlsnpl",
    "Kbdsmo": "Mlsnpi",
    "Di": "Sn",
    "Dr": "Sg",
    "Dd": "Sd",
    "Dt": "Sa",
    "Dm": "Sl",
    "Do": "Si",
    "Vp": "Cc",
    "Vd": "Cs",
    "L": "Q",
    "M": "I",
    "O": "Y",
    "N": "X",
    "Nj": "Xf",
    "Nt": "Xt",
    "Np": "Xp",
    "Nw": "Xw",
    "Ne": "Ne",
    "Nh": "Nh",
    "Na": "Na",
    "U": "Z"}



"""
Various progress logs.
"""
import time
import logging

try:
    from tqdm import tqdm
except ImportError:
    tqdm = None


REPORT_ON = 0.3


class Progress:
    def __call__(self, iterable, description, total=None):
        if tqdm is None:
            try:
                total = len(iterable)
            except TypeError:
                total = -1

            start_time = time.time()
            last_report = start_time - REPORT_ON
            for n, el in enumerate(iterable):
                now = time.time()
                if now - last_report > REPORT_ON:
                    logging.info("\r{}: {}/{}".format(description, n, total))
                    last_report = now
                yield el
            logging.info(" -> {}".format(time.time() - start_time))
        else:
            yield from tqdm(iterable, desc=description, total=total)


progress = Progress()



"""
Class for assigning representations.
"""

from cordex.representations.representation import ComponentRepresentation, LemmaCR, LexisCR, WordFormAgreementCR, WordFormAnyCR, WordFormMsdCR, WordFormAllCR


class RepresentationAssigner:
    def __init__(self):
        self.more = {}
        self.representation_factory = ComponentRepresentation

    def add_feature(self, feature):
        """ Sets up a proper representation factory, depending on feature. """
        if 'rendition' in feature:
            if feature['rendition'] == "lemma":
                self.representation_factory = LemmaCR
            elif feature['rendition'] == "word_form":
                # just by default, changes with selection
                self.representation_factory = WordFormAnyCR
            elif feature['rendition'] == "lexis":
                self.representation_factory = LexisCR
                self.more['lexis'] = feature['string']
            else:
                raise NotImplementedError("Representation rendition: {}".format(feature))

        elif 'selection' in feature:
            if feature['selection'] == "msd":
                # could already be agreement
                if self.representation_factory != WordFormAgreementCR:
                    self.representation_factory = WordFormMsdCR
                self.more['msd'] = {k: v for k, v in feature.items() if k != 'selection'}
            elif feature['selection'] == "all":
                self.representation_factory = WordFormAllCR
            elif feature['selection'] == 'agreement':
                assert feature['msd'] is not None
                self.representation_factory = WordFormAgreementCR
                self.more['agreement'] = feature['msd'].split('+')
                self.more['other'] = feature['head_cid']
            else:
                raise NotImplementedError("Representation selection: {}".format(feature))
        elif 'format' in feature:
            self.more['format'] = feature['format']

    def create_component_representation(self, word_renderer):
        """ Creates component representation. """
        return self.representation_factory(self.more, word_renderer)

    @staticmethod
    def set_representations(match, word_renderer, is_ud, representations, lookup_lexicon=None, lookup_api=False):
        """ Assigns representations to words. """
        for c in match.structure.components:
            representations[c.idx] = []
            for rep in c.representation:
                representations[c.idx].append(rep.create_component_representation(word_renderer))

        # links agreement representations with representations
        for cid, reps in representations.items():
            for rep in reps:
                for agr in rep.get_agreement_head_component_id():
                    if len(representations[agr]) != 1:
                        n = len(representations[agr])
                        raise NotImplementedError(
                            "Structure {}: ".format(match.structure.id) +
                            "component {} has agreement".format(cid) +
                            " with component {}".format(agr) +
                            ", however there are {} (!= 1) representations".format(n) +
                            " of component {}!".format(agr))

                    representations[agr][0].agreement.append(rep)

        # links matches with representations
        for words in match.matches:
            # first pass, check everything but agreements
            for w_id, w in words.items():
                component_representations = representations[w_id]
                for representation in component_representations:
                    representation.add_word(w, is_ud)

        for cid, reps in representations.items():
            for rep in reps:
                rep.render(is_ud, lookup_lexicon=lookup_lexicon, lookup_api=lookup_api)

        for cid, reps in representations.items():
            reps_text = [rep.rendition_text for rep in reps]
            reps_msd = [rep.rendition_msd for rep in reps]
            if reps_text == []:
                pass
            elif all(r is None for r in reps_text):
                match.representations[cid] = (None, None)
            else:
                reps_text_joined = " ".join(("" if r is None else r) for r in reps_text)
                reps_msd_joined = " ".join(("" if r is None else r) for r in reps_msd)
                match.representations[cid] = (reps_text_joined, reps_msd_joined)



"""
Tools for handling memory/actual database.
"""
import sqlite3
import os


class Database:
    def __init__(self, args):
        filename = ":memory:" if args['db'] is None else args['db']

        if args['overwrite_db'] and os.path.exists(filename):
            os.remove(filename)

        self.new = not os.path.exists(filename)
        self.db = sqlite3.connect(filename)

        self.init("CREATE TABLE StepsDone ( step varchar(32) )")
        self.commit()
    
    def execute(self, *args, **kwargs):
        """ Executes database command.  """
        return self.db.execute(*args, **kwargs)

    def init(self, *args, **kwargs):
        """ Same as execute, only skipped if not a new database file. """
        if self.new:
            return self.execute(*args, **kwargs)
    
    def commit(self):
        """ Commits changes. """
        self.db.commit()

    def is_step_done(self, step_name):
        """ Checks whether step results are in database. """
        wc_done = self.db.execute("SELECT count(*) FROM StepsDone WHERE step=?", (step_name, )).fetchone()
        return wc_done[0] != 0

    def step_is_done(self, step_name):
        """ Completes and stores step. """
        self.db.execute("INSERT INTO StepsDone (step) VALUES (?)", (step_name, ))
        self.db.commit()




"""
A class for storing matches.
"""
import gc
from collections import defaultdict
from ast import literal_eval
from time import time
import logging

from cordex.matcher.match import StructureMatch
from cordex.representations.representation_assigner import RepresentationAssigner
from cordex.utils.progress_bar import progress


class MatchStore:
    def __init__(self, args, db):
        self.db = db
        self.dispersions = {}
        self.min_freq = args['min_freq']
        self.is_ud = args['is_ud']

        # create necessary tables
        self.db.init("""CREATE TABLE Collocations (
            collocation_id INTEGER PRIMARY KEY,
            structure_id varchar(8),
            key varchar(256))
            """)
        if self.is_ud:
            self.db.init("""CREATE TABLE Matches (
                match_id INTEGER,
                component_id INTEGER NOT NULL,
                word_lemma varchar(32) NOT NULL,
                word_id varchar(32) NOT NULL,
                sentence_id varchar(32) NOT NULL,
                word_udpos varchar(32) NOT NULL,
                word_text varchar(32) NOT NULL)
                """)
        else:
            self.db.init("""CREATE TABLE Matches (
                            match_id INTEGER,
                            component_id INTEGER NOT NULL,
                            word_lemma varchar(32) NOT NULL,
                            word_id varchar(32) NOT NULL,
                            sentence_id varchar(32) NOT NULL,
                            word_xpos varchar(16) NOT NULL,
                            word_text varchar(32) NOT NULL)
                            """)
        self.db.init("""CREATE TABLE CollocationMatches (
            mid_match_id INTEGER,
            mid_collocation_id INTEGER,
            FOREIGN KEY(mid_collocation_id) REFERENCES Collocations(collocation_id),
            FOREIGN KEY(mid_match_id) REFERENCES Matches(match_id))
            """)
        self.db.init("""CREATE TABLE Representations (
            collocation_id INTEGER,
            component_id INTEGER,
            text varchar(32),
            msd varchar(32),
            FOREIGN KEY(collocation_id) REFERENCES Collocations(collocation_id))
            """)
        self.db.init("""CREATE TABLE Dispersions (
            structure_id varchar(64),
            component_id varchar(64),
            lemma varchar(128),
            dispersion INTEGER)
            """)
        
        self.db.init("CREATE INDEX key_sid_c ON Collocations (key, structure_id)")
        self.db.init("CREATE INDEX sid_c ON Collocations (structure_id)")
        self.db.init("CREATE INDEX mmid_cm ON CollocationMatches (mid_collocation_id)")
        self.db.init("CREATE INDEX mid_m ON Matches (match_id)")
        self.db.init("CREATE INDEX col_r ON Representations (collocation_id)")
        self.db.init("CREATE INDEX disp_key ON Dispersions (structure_id, component_id, lemma)")

        match_num = self.db.execute("SELECT MAX(match_id) FROM Matches").fetchone()[0]
        self.match_num = 0 if match_num is None else match_num + 1

    def _add_match(self, key, match):
        """ Inserts a match to database. """
        structure_id, key_str = key[0], str(key[1:])
        cid = self.db.execute("SELECT collocation_id FROM Collocations WHERE key=? AND structure_id=?",
                              (key_str, structure_id)).fetchone()

        if cid is None:
            self.db.execute("INSERT INTO Collocations (structure_id, key) VALUES (?,?)",
                            (structure_id, key_str))
            cid = self.db.execute("SELECT collocation_id FROM Collocations WHERE key=? AND structure_id=?",
                                  (key_str, structure_id)).fetchone()
        
        for component_id, word in match.items():
            if self.is_ud:
                self.db.execute("""
                INSERT INTO Matches (match_id, component_id, word_lemma, word_text, word_udpos, word_id, sentence_id) 
                VALUES (:match_id, :component_id, :word_lemma, :word_text, :word_udpos, :word_id, :sentence_id)""", {
                    "component_id": component_id,
                    "match_id": self.match_num,
                    "word_lemma": word.lemma,
                    "word_udpos": str(word.udpos),
                    "word_text": word.text,
                    "word_id": word.id,
                    "sentence_id": word.sentence_id,
                })
            else:
                self.db.execute("""
                INSERT INTO Matches (match_id, component_id, word_lemma, word_text, word_xpos, word_id, sentence_id) 
                VALUES (:match_id, :component_id, :word_lemma, :word_text, :word_xpos, :word_id, :sentence_id)""", {
                    "component_id": component_id,
                    "match_id": self.match_num,
                    "word_lemma": word.lemma,
                    "word_xpos": word.xpos,
                    "word_text": word.text,
                    "word_id": word.id,
                    "sentence_id": word.sentence_id,
                })

        self.db.execute("INSERT INTO CollocationMatches (mid_collocation_id, mid_match_id) VALUES (?,?)",
                        (cid[0], self.match_num))
        
        self.match_num += 1

    def add_matches(self, matches):
        """ Add multiple matches. """
        for structure, nms in progress(matches.items(), 'adding-matches'):
            for nm in nms:
                self._add_match(nm[1], nm[0])

    def get_matches_for(self, structure):
        """ Get all matches for given structure. """
        for cid in self.db.execute("SELECT collocation_id FROM Collocations WHERE structure_id=?",
                                   (structure.id,)):
            yield StructureMatch.from_db(self.db, cid[0], structure, self.is_ud)

    def add_inserts(self, inserts):
        """ Adds representations to database. """

        for match in inserts:
            for component_id, (text, msd) in match.representations.items():
                self.db.execute("""
                    INSERT INTO Representations (collocation_id, component_id, text, msd) 
                    VALUES (?,?,?,?)""", (match.match_id, component_id, text, msd))

    def set_representations(self, word_renderer, structures, is_ud, lookup_lexicon=None, lookup_api=False):
        """ Adds representations to matches. """

        step_name = 'representation'
        if self.db.is_step_done(step_name):
            logging.info("Representation step already done, skipping")
            return

        num_inserts = 1000
        inserts = []

        structures_dict = {s.id: s for s in structures}
        num_representations = int(self.db.execute("SELECT Count(*) FROM Collocations").fetchone()[0])

        # preprocess when api is used
        all_representations = []
        if lookup_api:
            # create api queries
            for cid, sid in progress(self.db.execute("SELECT collocation_id, structure_id FROM Collocations"),
                                     "representations", total=num_representations):
                structure = structures_dict[sid]
                match = StructureMatch.from_db(self.db, cid, structure, is_ud)
                representations = {}
                RepresentationAssigner.set_representations(match, word_renderer, is_ud, representations,
                                                           lookup_lexicon=lookup_lexicon, lookup_api=lookup_api)

                all_representations.append(representations)

            # get data results
            lookup_api.execute_requests(num_representations, self.db, all_representations)

        start_time = time()
        i = 0
        for cid, sid in progress(self.db.execute("SELECT collocation_id, structure_id FROM Collocations"), "representations", total=num_representations):
            structure = structures_dict[sid]
            match = StructureMatch.from_db(self.db, cid, structure, is_ud)
            representations = {}
            RepresentationAssigner.set_representations(match, word_renderer, is_ud, representations, lookup_lexicon=lookup_lexicon, lookup_api=lookup_api)

            inserts.append(match)
            if len(inserts) > num_inserts:
                self.add_inserts(inserts)
                inserts = []
            if time() - start_time > 5:
                start_time = time()
                gc.collect()
            if lookup_api:
                all_representations.append(representations)
            i += 1
        self.add_inserts(inserts)

        self.db.step_is_done(step_name)

    def frequency_filter(self, collocation_id):
        """ Filters by frequency. """

        matches = self.db.execute("SELECT MIN(MAX(COUNT(*), ?), ?) FROM CollocationMatches WHERE mid_collocation_id=?", (self.min_freq - 1, self.min_freq, collocation_id)).fetchone()[0]
        return matches >= self.min_freq

    def determine_collocation_dispersions(self):
        """ Allocates collocation dispersions. """
        step_name = 'dispersions'
        if self.db.is_step_done(step_name):
            self.load_dispersions()
            return

        dispersions = defaultdict(int)
        for collocation_id, structure_id, word_tups_str in progress(self.db.execute("SELECT collocation_id, structure_id, key FROM Collocations"), "dispersion"):
            if not self.frequency_filter(collocation_id):
                continue

            word_tups = literal_eval(word_tups_str)
            for component_id, lemma in word_tups:
                dispersions[(str(structure_id), component_id, lemma)] += 1
            
        self.dispersions = dict(dispersions)
        logging.info("Storing dispersions...")
        self.store_dispersions()

        self.db.step_is_done(step_name)

    def store_dispersions(self):
        """ Stores dispersions in sql database. """
        for (structure_id, component_id, lemma), disp in self.dispersions.items():
            self.db.execute("INSERT INTO Dispersions (structure_id, component_id, lemma, dispersion) VALUES (?, ?, ?, ?)",
                            (structure_id, component_id, lemma, disp))

    def load_dispersions(self):
        """ Load dispersions when they are in database. """
        self.dispersions = {}
        for structure_id, component_id, lemma, dispersion in progress(self.db.execute("SELECT * FROM Dispersions"), "load-dispersions"):
            self.dispersions[structure_id, component_id, lemma] = dispersion
